/*
	File:		Animation.cc

	Function:	

	Author:		Andrew Willmott

	Notes:		
*/

#include "gcl/Animation.h"
// XXX
#include "GCLApp.h"
#include "gcl/GCLConfig.h"
 
#define DBG_COUT if (0) cerr


Void AnimRec::AddKey(GCLReal time, GCLReal value, GCLReal vIn, GCLReal vOut)
{
	keyTimes.Append(time);
	values.Append(value);
	velIn.Append(vIn);
	velOut.Append(vOut);
}

Void AnimRec::SetupAnim()
{
	Int		i, n = keyTimes.NumItems();
	GCLReal	autoVel;
	
	// fill in automatic velocities

	if (n <= 1)
		return;

	autoVel = (values[1] - values[0]) / (keyTimes[1] - keyTimes[0]);
	if (velIn[0] == kAutoVel)
		velIn[0] = autoVel;
	if (velOut[0] == kAutoVel)
		velOut[0] = autoVel;
	
	for (i = 1; i < keyTimes.NumItems() - 1; i++)
	{
		autoVel = (values[i + 1] - values[i - 1]) / (keyTimes[i + 1] - keyTimes[i - 1]);

		if (velIn[i] == kAutoVel)
			velIn[i] = autoVel;
		if (velOut[i] == kAutoVel)
			velOut[i] = autoVel;
	}

	autoVel = (values[n - 1] - values[n - 2]) / (keyTimes[n - 1] - keyTimes[n - 2]);
	if (velIn[n - 1] == kAutoVel)
		velIn[n - 1] = autoVel;
	if (velOut[n - 1] == kAutoVel)
		velOut[n - 1] = autoVel;
}

Int AnimRec::FindTimeIndex(GCLReal time)
{
	Int 	i = 0;
	
	if (currentTimeIndex >= 0 && time > keyTimes[currentTimeIndex])
		i = currentTimeIndex + 1;
	while (i < keyTimes.NumItems() && time > keyTimes[i])
		i++;
	
	return(i - 1);
}

GCLReal AnimRec::GetValue(GCLReal time)
{
	Int 		i;
	GCLReal 	t1, t2, y1, y2, s1, s2;
	
	Assert(keyTimes.NumItems() > 0, "(AnimRec::GetValue) must have at least one keyframe");

	i = FindTimeIndex(time);

	if (i < 0)
		return(values[0]);
	if (i >= keyTimes.NumItems() - 1)
		return(values[keyTimes.NumItems() - 1]);

	t1 = keyTimes[i];
	t2 = keyTimes[i + 1];

	if (i != currentTimeIndex)
	{
		// recalculate cubic for next bracket
		DBG_COUT << "new cubic for t = " << time << endl;
		currentTimeIndex = i;
		y1 = values[i];
		y2 = values[i + 1];
			 
#ifdef OLD
		if (i == 0)
			// no prior keyframe
			s1 = (y2 - y1) / (t2 - t1);
		else
			s1 = (y2 - values[i - 1]) / (t2 - keyTimes[i - 1]);

		if (i == keyTimes.NumItems() - 2)
			// no posterior keyframe
			s2 = (y2 - y1) / (t2 - t1);
		else
			s2 = (values[i + 2] - y1) / (keyTimes[i + 2] - t1);
#else
		s1 = velOut[i];
		s2 = velIn[i + 1];
#endif
			
		CalcHermiteCoeffs(t1, t2, y1, y2, s1, s2, currentCubic);
	}
	
	return(EvalCubic(currentCubic, time, t1, t2));
}


// --- Animation class --------------------------------------------------------


struct Dumper : IntHashIter
{
	Dumper(AnimRecList &a, GCLReal t, Bool ad) : 
		recs(a), time(t), addDash(ad) {};

	Void	ProcessItem(StrConst s, Int a)	
	{
		cout << ' ';
		if (addDash)
			cout << '-';
		cout << s << ' ' << recs[a]->GetValue(time);
	}

	AnimRecList			&recs;
	GCLReal				time;
	Bool				addDash;
};

Void Animation::DumpAvars(GCLReal time, Int frame)
{
	Dumper	dumper(avarRecs, time, addDash);

	if (!quiet)
	{
		if (tag)
			cout << tag << '_' << frame;
		else
			cout << "time " << time << " frame " << frame;
	}
	avarNames.Iterate(dumper);
	cout << endl;
}

GCLReal Animation::GetValue(StrConst avarName, GCLReal time)
{
	return(avarRecs[avarNames.GetItem(avarName)]->GetValue(time));
}

struct CameraUpdater : IntHashIter
{
	CameraUpdater(AnimRecList &a, GCLReal t, Camera &c) : 
		recs(a), time(t), camera(c) {};

	Void ProcessItem(StrConst s, Int a)	
	{
		camera.CheckAvar(s, recs[a]->GetValue(time));
	}

	AnimRecList			&recs;
	Camera				&camera;
	GCLReal				time;
};

Void Animation::SetAvarList(scAvarList *avarList, scCamera *itsCamera, GCLReal time)
{
	Int				i;

	if (avarMap.NumItems() == 0 && avarList->NumItems() > 0)
	{
		// build avarMap: maps from avarList entry to to avarRecs entry

		avarMap.SetSize(avarList->NumItems());
		for (i = 0; i < avarList->NumItems(); i++)
			avarMap[i] = avarNames.GetItem((*avarList)[i].name);
	}
	
	// update avar list
	for (i = 0; i < avarList->NumItems(); i++)
		if (avarMap[i] >= 0)
			(*avarList)[i].value = avarRecs[avarMap[i]]->GetValue(time);

	if (itsCamera)
	{
		CameraUpdater cu(avarRecs, time, *itsCamera);

		avarNames.Iterate(cu);

		itsCamera->SetupFromParams();
	}
}

Void Animation::AddKey(StrConst avarName, GCLReal time, GCLReal value,
	GCLReal velIn, GCLReal velOut)
{
	Int		i;
	
	if (!avarNames.ItemExists(avarName))
	{
		// new avar: create avarRec

		i = avarRecs.NumItems();
		avarNames.SetItem(avarName, i);
		avarRecs.Append(new AnimRec);
	}
	else
		i = avarNames.GetItem(avarName);

	avarRecs[i]->AddKey(time, value, velIn, velOut);
}

/** Parse animation file.

	\verbatim
	Expected format for parser: wavefront-like

	t <time>
		set time: expected to be monotonic!
	dt <delta time>
		set time as delta from last set time.
	v|vi|vo <avar name> [<velocity> | auto]
		set velocity in, velocity out, or both at once. the auto keyword
		means calculate the velocity automatically.
	k <avar name> <value>
		set avar value at the current time with current velocities
	\endverbatim
*/

Bool Animation::ParseAnimFile(StrConst filename)
{
	String		func;
	ifstream 	s;
	
	Int			lines = 0;
	GCLReal		currentTime = 0.0;
	String		avarName, avarVal;
	GCLReal		value = 0.0;
	GCLReal		velIn = kAutoVel;
	GCLReal		velOut = kAutoVel;
	GCLReal		setVel;
	
	length = 0.0;
	avarRecs.Clear();

	s.open(filename);
	if (!s)
	{
		cerr << "Cannot access " << filename << endl;
		return(false);
	}

	while (s)
	{
		if (func.ReadWord(s))
		{
			lines++;
			if (func[0] == '#')
				;
			else if (func == "t")
			{
				s >> currentTime;
				length = Max(length, currentTime);
				// reset velocities to auto
				velIn = velOut = kAutoVel;
			}
			else if (func == "dt")
			{
				GCLReal delta;

				s >> delta;
				currentTime += delta;
				length = Max(length, currentTime);
				// reset velocities to auto
				velIn = velOut = kAutoVel;
			}
			else if (func == "k")
			{
				avarName.ReadWord(s);
				
				if (avarName.Suffix(2) == "_q")
				{
					String stub = avarName.Prefix(-2);

					avarName.Printf("%s_x", stub.CString());
					s >> value;
					AddKey(avarName, currentTime, value, velIn, velOut);
					
					avarName.Printf("%s_y", stub.CString());
					s >> value;
					AddKey(avarName, currentTime, value, velIn, velOut);

					avarName.Printf("%s_z", stub.CString());
					s >> value;
					AddKey(avarName, currentTime, value, velIn, velOut);

					avarName.Printf("%s_w", stub.CString());
					s >> value;
					AddKey(avarName, currentTime, value, velIn, velOut);
				}
				else if (avarName.Suffix(2) == "_v")
				{
					String stub = avarName.Prefix(-2);

					avarName.Printf("%s_x", stub.CString());
					s >> value;
					AddKey(avarName, currentTime, value, velIn, velOut);
					
					avarName.Printf("%s_y", stub.CString());
					s >> value;
					AddKey(avarName, currentTime, value, velIn, velOut);

					avarName.Printf("%s_z", stub.CString());
					s >> value;
					AddKey(avarName, currentTime, value, velIn, velOut);
				}
				else if (avarName.Suffix(2) == "_c")
				{
					String stub = avarName.Prefix(-2);

					avarName.Printf("%s_x", stub.CString());
					s >> value;
					AddKey(avarName, currentTime, value, velIn, velOut);
					
					avarName.Printf("%s_y", stub.CString());
					s >> value;
					AddKey(avarName, currentTime, value, velIn, velOut);
				}
				else
				{
					s >> value;
					AddKey(avarName, currentTime, value, velIn, velOut);
				}
			}
			else if (func[0] == 'v')
			{
				avarVal.ReadWord(s);
				if (avarVal == "auto")
					setVel = kAutoVel;
				else
					setVel = atof(avarVal.CString());

				if (func == "v")
					velOut = velIn = setVel;
				else if (func == "vi")
					velIn = setVel;
				else if (func == "vo")
					velOut = setVel;
			}
			else
				cerr << "(ParseAnimFile) *** Ignoring unknown token: " << func << endl;

			func.ReadLine(s); // ignore rest of line
		}
	}
		
	SetupAnim();
	
	return(true);
}

Void Animation::SetupAnim()
{
	for (Int i = 0; i < avarRecs.NumItems(); i++)
		avarRecs[i]->SetupAnim();
}
