/*
	File:		ImageComp.cc

	Function:	Compare two image files

	Author:		Andrew Willmott

	Notes:		
*/

#include "cl/ArgParse.h"
#include <stdlib.h>
#include <unistd.h>

#include "gcl/Image.h"
#include "gcl/XGraphicsSystem.h"
#include "cl/Action.h"

Void ApplyToImageDiff(Action<ClrReal> &action, Image &image1, Image &image2,
			ImgChannel chn = chMono)
{
	Int		i, j;
	ClrReal	temp;

	action.Start();
	
	for (i = 0; i < image1.Height(); i++)
		for (j = 0; j < image1.Width(); j++)
		{
			temp = image1.GetRealPixel(j, i, chn) - image2.GetRealPixel(j, i, chn);
			action.Process(temp);
		}	

	action.Stop();
}

ClrReal AbsImageDiff(Image &image1, Image &image2, Image &destImage)
{
	Int		i, j;
	Colour	c;
	ClrReal	maxCmpt = 0.0;
	
	for (i = 0; i < image1.Height(); i++)
		for (j = 0; j < image1.Width(); j++)
		{
			Colour a = image1.GetPixel(j, i);
			Colour b = image2.GetPixel(j, i);

			c = a - b;
			c[0] = abs(c[0]);
			c[1] = abs(c[1]);
			c[2] = abs(c[2]);

			maxCmpt = Max(maxCmpt, Max(c[0], Max(c[1], c[2])));
			destImage.SetPixel(j, i, c);
		}

	return(maxCmpt);
}


// --- Main Program -----------------------------------------------------------


Int main(Int argc, Char **argv)
{
	XGraphicsSystem         *gs;       
	StatsFinder<ClrReal>    rstats, bstats, gstats;     
	RGBAImage               img1, img2, imgDiff;     
	XOffscreenPane          *offImg;       
	XBackedPane             *window;       
	FileName                fpath;       
	Char                    *file1, *file2;      
	Int                     useWin, noFork;
	ClrReal					maxCmpt;
	
	// command-line options
	
	if (arg_parse(argc, argv, 
			"", 										"usage: prog [options]",
			
			"%S",	&file1,								"image file 1",
			"%S",	&file2,								"image file 2",
			
			// Output options
			
			"-window", ARG_FLAG(&useWin), 				"display image differences in window",
			"-noFork", ARG_FLAG(&noFork),				"Don't fork into background",
			0) < 0)
		exit(1);
	
	// Read two images
	
	fpath.SetPath(file1);
	if (img1.Load(fpath) != 0)
		exit(1);
	
	if (useWin)
	{
#ifndef DEBUG
		if (!noFork && fork())
			return(0);
#endif
		gs = new XGraphicsSystem;
		offImg = new XOffscreenPane;
		gs->CreateOffscreen(offImg, img1.Width(), img1.Height());
		window = new XBackedPane;
		gs->CreateWindow(window, "difference", img1.Width(), img1.Height());
		offImg->PutImage(img1);
		window->Attach(offImg);
		window->HandleExpose();
		gs->Spin();
	}
	
	fpath.SetPath(file2);
	if (img2.Load(fpath) != 0)
		exit(1);
	
	if (img2.Width() != img1.Width() || img2.Height() != img1.Height())
	{
		cerr << "Error: picture dimensions don't match." << endl;
		exit(1);
	}
	
	if (useWin)
	{
		offImg->PutImage(img2);
		window->HandleExpose();
		gs->Spin();
	}
		
	// Compare...
	
	ApplyToImageDiff(rstats, img1, img2, chRed);
	ApplyToImageDiff(gstats, img1, img2, chGreen);
	ApplyToImageDiff(bstats, img1, img2, chBlue);
	
	cout << "mean_R " << rstats.mean << endl;
	cout << "var_R " << rstats.variance << endl;
	cout << "dev_R " << sqrt(rstats.variance) << endl;
	cout << endl;
	cout << "mean_G " << gstats.mean << endl;
	cout << "var_G " << gstats.variance << endl;
	cout << "dev_G " << sqrt(gstats.variance) << endl;
	cout << endl;
	cout << "mean_B " << bstats.mean << endl;
	cout << "var_B " << bstats.variance << endl;
	cout << "dev_B " << sqrt(bstats.variance) << endl;
	cout << endl;
	cout << "mean_L " << dot(cRGBToLum, Colour(rstats.mean, gstats.mean, bstats.mean)) << endl;
	cout << "var_L " << dot(cRGBToLum, Colour(rstats.variance, gstats.variance, bstats.variance)) << endl;
	cout << "dev_L " << sqrt(dot(cRGBToLum, Colour(rstats.variance, gstats.variance, bstats.variance))) << endl;
	
	if (useWin)
	{
		imgDiff.SetSize(img1.Width(), img1.Height());
		maxCmpt = AbsImageDiff(img1, img2, imgDiff);	
		imgDiff.Scale(1.0 / maxCmpt);
		
		offImg->PutImage(imgDiff);
		window->HandleExpose();
		gs->Run();
	}
	
	return(0);
}
