/*
	File:		ParseMGF.cc

	Function:	Provides routine to parse .mgf files into a scene.

	Author:		Andrew Willmott

	Notes:		Need to support shared vertices, vertex normals.

	Copyright:	(c) 1996-2000, Andrew Willmott
*/

#include "gcl/GCLConfig.h"
#ifdef GCL_MGF
#include "gcl/SceneLang.h"
#include <stdio.h>
#include <string.h>

extern "C" 
{
#include <string.h>
#include "parser.h"
#include "lookup.h"
}

static int myObject(int ac, char **av);
static int myTransform(int ac, char **av);
static int myCylinder(int ac, char **av);
static int myFace(int ac, char **av);
static int mySphere(int ac, char **av);
static int myComment(int ac, char **av);

#ifdef DEBUG
#define DBG_COUT if (0) cerr
//#define MGF_DUMP
#else
#define DBG_COUT if (0) cerr
#endif

Colour MGF2Colour(
		C_COLOR		*cin,           // input MGF chrominance
		Double		intensity       // input luminance or reflectance 
	);

/*!
	\method ParseMGFFile
	
	See http://radsite.lbl.gov/mgf/HOME.html.
*/

scScenePtr ParseMGFFile(const Char *filename)
{
	scScenePtr result;

	result = slBeginObject(filename);
	
	// rotate from MGF's standard zup to our standard yup
	slTransform(Rotation(vl_x, -vl_halfPi));
	// initialize dispatch table 

	mg_ehand[MG_E_COMMENT] = myComment;	
	mg_ehand[MG_E_COLOR] = c_hcolor;	
	mg_ehand[MG_E_CMIX] = c_hcolor;		
	mg_ehand[MG_E_CSPEC] = c_hcolor;	
	mg_ehand[MG_E_CXY] = c_hcolor;		
	mg_ehand[MG_E_CCT] = c_hcolor;		
	//mg_ehand[MG_E_CYL] = myCylinder;	
	mg_ehand[MG_E_ED] = c_hmaterial;	
	mg_ehand[MG_E_FACE] = myFace;		
	mg_ehand[MG_E_MATERIAL] = c_hmaterial;
	mg_ehand[MG_E_NORMAL] = c_hvertex;	
	//mg_ehand[MG_E_OBJECT] = myObject;	
	mg_ehand[MG_E_POINT] = c_hvertex;	
	mg_ehand[MG_E_RD] = c_hmaterial;	
	mg_ehand[MG_E_RS] = c_hmaterial;	
	mg_ehand[MG_E_SIDES] = c_hmaterial;	
	//mg_ehand[MG_E_SPH] = mySphere;	
	mg_ehand[MG_E_TD] = c_hmaterial;	
	mg_ehand[MG_E_TS] = c_hmaterial;	
	mg_ehand[MG_E_VERTEX] = c_hvertex;	
	mg_ehand[MG_E_XF] = xf_handler;		

	DBG_COUT << "initialising parser" << endl;
	
	mg_init();							/* initialize the parser */
	
	if (mg_load((char*)filename) != MG_OK)
	{
		slEndObject();
		return(0);
	}
	slEndObject();
	return(result);
}


static Int myObject(int ac, char **av)			/* group object name */
{
	static int	objnest;

#ifdef MGF_DUMP
	int i;
	for (i = 0; i < ac; i++)
		printf("%s ", av[i]);
	printf("\n");
#endif

	return(MG_OK);
}

static int myComment(int ac, char **av)
{
#ifdef MGF_DUMP
	int i;
	printf("comment:\n");
	for (i = 0; i < ac; i++)
		printf("%s ", av[i]);
	printf("\n");
#endif
	return(MG_OK);
}

#if 0
// shared vertex lookup table
const Char *kVertFmt 	= "%+16.9e %+16.9e %+16.9e %+6.3f %+6.3f %+6.3f";
const Char *kVZVect 	= "+0.000 +0.000 +0.000";
const Int kVFLen 		= 72;
const Int kMaxVert		10240;	/* maximum cached vertices */

#define setvkey(k,v)	sprintf(k,VERTFMT,(v)->p[0],(v)->p[1],(v)->p[2],\
					(v)->n[0],(v)->n[1],(v)->n[2]);

char	vlist[MAXVERT][VFLEN];	/* our vertex cache */
int	nverts;			/* current cache size */

LUTAB	vert_tab = LU_SINIT(NULL,NULL);

#endif


static int myFace(int ac, char **av)			/* translate an N-sided face */
{
	C_VERTEX	*vp;
	int 		i;
	FVECT		n, p;
	Bool		doNormals = false, ctxChange = false;
	Int			idx;
	
	static Int			gLastXID = -1;
	static Colour		gLastRClr(-1, -1, -1), gLastEClr(-1, -1, -1);

	if (ac < 4)
		return(MG_EARGC);
	
	if (xf_context && xf_context->xid != gLastXID)
	{
		gLastXID = xf_context->xid;
		ctxChange = true;
		DBG_COUT << "transform context changed: rev = " << xf_context->rev << endl;
	}
		
	DBG_COUT << "mtl: " << c_cmname << ' ' << c_cmaterial << ' ' << c_cmaterial->clock << endl;
	if (c_cmaterial->clock != 0)
	// material has changed...
	{
		Colour	rClr, eClr;
		
		DBG_COUT << "material change: clock = " << c_cmaterial->clock << endl;
		
		rClr = MGF2Colour(&c_cmaterial->rd_c, c_cmaterial->rd);
		eClr = MGF2Colour(&c_cmaterial->ed_c, c_cmaterial->ed);
		// the MGF 'ed' is measured in lux, photometric irradiance.
		// convert to our pure irradiance
		eClr /= 179.0;

		if (rClr != gLastRClr)
		{
			DBG_COUT << "reflectance = " << rClr << endl;
			slColour(rClr);
			gLastRClr = rClr;
			ctxChange = true;
		}
		if (eClr != gLastEClr)
		{
			DBG_COUT << "emittance = " << eClr << endl;
			slEmittance(eClr);
			gLastEClr = eClr;
			ctxChange = true;
		}

		c_cmaterial->clock = 0;
	}
		
#ifdef MGF_DUMP
	printf("face:\n");
	for (i = 0; i < ac; i++)
		printf("%s ", av[i]);
	printf("\n");
#endif

	slPointList();

	for (i = 1; i < ac; i++) 
	{
		if (xf_context && xf_context->rev) 
			vp = c_getvert(av[ac - i]);
		else
			vp = c_getvert(av[i]);
		
		if (vp == NULL)
			return(MG_EUNDEF);
		
		// we wish to use shared vertices -- it appears the only
		// way to do this is to use a hash =P

#ifdef MGF_SHARED_VERTS
		// XXX unfinished!
		lu_init(&vert_tab, MAXVERT);

		setvkey(vlist[nverts], vp);
		lp = lu_find(&vert_tab, vlist[nverts]);
		if (lp == NULL)
			return(MG_EMEM);
		if (lp->key == NULL)
			lp->key = (char *)vlist[nverts++];
		newf->vl[i] = ((char (*)[VFLEN])lp->key - vlist);

		idx = (Int) vp->client_data;

		if (ctxChange || vp->clock != 0 || idx == 0)
		{
			vp->clock = 0;
			xf_xfmpoint(p, vp->p);
			idx = slPoint(Vector(p[0], p[1], p[2]));
			vp->client_data = (char*) (idx + 1);
			DBG_COUT << "create index = " << idx << ", " << vp << endl;

			if (i == 1 && (vp->n[0] != 0.0 || vp->n[1] != 0.0 || vp->n[2] != 0.0))
			{
				// if the first vertex has a vertex normal defined, assume the
				// rest do too.
				slNormalList();
				doNormals = true;
			}

			if (doNormals)
			{
				Vector	nrm;

				xf_rotvect(n, vp->n);
				nrm = Vector(n[0], n[1], n[2]);
				DBG_COUT << "norm = " << nrm << endl;
				nrm.Normalise();
				slNormal(nrm);
			}
		}
		else
		{
			idx = idx - 1;
			DBG_COUT << "reused index = " << idx << ", " << vp << endl;
		}
		//slIndex(idx);
	}
#else
		xf_xfmpoint(p, vp->p);
		slPoint(Vector(p[0], p[1], p[2]));

		if (i == 1 && (vp->n[0] != 0.0 || vp->n[1] != 0.0 || vp->n[2] != 0.0))
		{
			// if the first vertex has a vertex normal defined, assume the
			// rest do too.
			slNormalList();
			doNormals = true;
		}

		if (doNormals)
		{
			Vector	nrm;

			xf_rotvect(n, vp->n);
			nrm = Vector(n[0], n[1], n[2]);
			DBG_COUT << "norm = " << nrm << endl;
			nrm.Normalise();
			slNormal(nrm);
		}
	}
	slPoly();
#endif
	
	return(MG_OK);			
}


static int mySphere(int ac, char **av)			/* translate sphere description */
{
#ifdef MGF_DUMP
	printf("sphere:\n");
#endif
	return(MG_OK);		/* we'll actually put it out later */
}


static int myCylinder(int ac, char **av)			/* translate a cylinder description */
{
#ifdef MGF_DUMP
	printf("cyl:\n");
	int i;
	for (i = 0; i < ac; i++)
		printf("%s ", av[i]);
	printf("\n");
#endif
	return(MG_OK);		/* we'll actually put it out later */
}


#ifndef OLD
                        /* Change the following to suit your standard */
#define  CIE_x_r                0.640           /* nominal CRT primaries */
#define  CIE_y_r                0.330
#define  CIE_x_g                0.290
#define  CIE_y_g                0.600
#define  CIE_x_b                0.150
#define  CIE_y_b                0.060
#define  CIE_x_w                0.3333          /* use true white */
#define  CIE_y_w                0.3333

#define CIE_D           (       CIE_x_r*(CIE_y_g - CIE_y_b) + \
                                CIE_x_g*(CIE_y_b - CIE_y_r) + \
                                CIE_x_b*(CIE_y_r - CIE_y_g)     )

#define CIE_C_rD        ( (1./CIE_y_w) * \
                                ( CIE_x_w*(CIE_y_g - CIE_y_b) - \
                                  CIE_y_w*(CIE_x_g - CIE_x_b) + \
                                  CIE_x_g*CIE_y_b - CIE_x_b*CIE_y_g     ) )
#define CIE_C_gD        ( (1./CIE_y_w) * \
                                ( CIE_x_w*(CIE_y_b - CIE_y_r) - \
                                  CIE_y_w*(CIE_x_b - CIE_x_r) - \
                                  CIE_x_r*CIE_y_b + CIE_x_b*CIE_y_r     ) )
#define CIE_C_bD        ( (1./CIE_y_w) * \
                                ( CIE_x_w*(CIE_y_r - CIE_y_g) - \
                                  CIE_y_w*(CIE_x_r - CIE_x_g) + \
                                  CIE_x_r*CIE_y_g - CIE_x_g*CIE_y_r     ) )

#define CIE_rf          (CIE_y_r*CIE_C_rD/CIE_D)
#define CIE_gf          (CIE_y_g*CIE_C_gD/CIE_D)
#define CIE_bf          (CIE_y_b*CIE_C_bD/CIE_D)

/* XYZ to RGB conversion matrix */

const ClrMat xyz2rgb
(
	 (CIE_y_g - CIE_y_b - CIE_x_b * CIE_y_g + CIE_y_b * CIE_x_g) / CIE_C_rD,
	 (CIE_x_b - CIE_x_g - CIE_x_b * CIE_y_g + CIE_x_g * CIE_y_b) / CIE_C_rD,
	 (CIE_x_g * CIE_y_b - CIE_x_b * CIE_y_g) / CIE_C_rD,

	 (CIE_y_b - CIE_y_r - CIE_y_b * CIE_x_r + CIE_y_r * CIE_x_b) / CIE_C_gD,
	 (CIE_x_r - CIE_x_b - CIE_x_r * CIE_y_b + CIE_x_b * CIE_y_r) / CIE_C_gD,
	 (CIE_x_b * CIE_y_r - CIE_x_r * CIE_y_b) / CIE_C_gD,

	 (CIE_y_r - CIE_y_g - CIE_y_r * CIE_x_g + CIE_y_g * CIE_x_r) / CIE_C_bD,
	 (CIE_x_g - CIE_x_r - CIE_x_g * CIE_y_r + CIE_x_r * CIE_y_g) / CIE_C_bD,
	 (CIE_x_r * CIE_y_g - CIE_x_g * CIE_y_r) / CIE_C_bD
);

// convert MGF color to RGB 

Colour	gMGFRGBLum(CIE_rf, CIE_gf, CIE_bf);

Colour MGF2Colour(
		C_COLOR		*cin,          /* input MGF chrominance */
		Double		intensity      /* input luminance or reflectance */
	)
{
	Colour	xyz, rgb;
	ClrReal rgbLum;
	
	// get CIE XYZ representation 
	c_ccvt(cin, C_CSXY);

	DBG_COUT << "converting: x y L: " << cin->cx << ' ' << cin->cy << ' ' << intensity << endl;

	xyz[0] = cin->cx;
	xyz[1] = cin->cy;
	xyz[2] = (1.0 - cin->cx - cin->cy);

	// convert to rgb colour

	// XXX change this to use ColourSystem stuff.
	rgb = xyz2rgb * xyz;
	// correct for out of gamut
	ClipColourZero(rgb);
	// normalise rgb luminance to 1
	rgbLum = (CIE_rf * rgb[0] + CIE_gf * rgb[1] + CIE_bf * rgb[2]);
	if (rgbLum > 0.0)
		rgb /= rgbLum;
	// set intensity.
	rgb *= intensity;

	DBG_COUT << "got: " << rgb << endl;

	return(rgb);
}
#else

#include "gcl/ColourSystem.h"

static ColourSystem mgfCS(
	Chroma(0.640, 0.330), 
	Chroma(0.290, 0.600), 
	Chroma(0.150, 0.060),
	Chroma(0.3333, 0.3333));

Colour MGF2Colour(
		C_COLOR		*cin,          /* input MGF chrominance */
		Double		intensity      /* input luminance or reflectance */
	)
{
	Colour	rgb;
	
	// get CIE XYZ representation 
	c_ccvt(cin, C_CSXY);

	DBG_COUT << "converting: x y L: " << cin->cx << ' ' << cin->cy << ' ' << intensity << endl;

	rgb = mgfCS.ChromaToGamut(Chroma(cin->cx, cin->cy));
	rgb *= intensity;
	
	DBG_COUT << "got: " << rgb << endl;

	return(rgb);
}


#endif

#endif
