/*
	File:			ScenePane.cc

	Function:		See header file

	Author(s):		Andrew Willmott

	Copyright:		(c) 1995-2000, Andrew Willmott

	Notes:			May have to split some of this stuff out.
					To come: picking, quaternion rotation (mouseball) &
					mouse pole.
*/

#include "gcl/GCLConfig.h"

#ifndef GCL_NO_GL

#include "gcl/ScenePane.h"
#include "gcl/GCLConfig.h"
#include "gcl/MRModel.h"
#include "gcl/SceneLang.h"

#ifdef VL_ROW_ORIENT
#error not_implemented
#endif


// --- Frame primitive ---------------------------------------------------------

// this is just for display -- don't add it to any model definitions.

class scFrame : public scPrimitive
{
public:
	scFrame() : 
		scPrimitive(pExtension),
		divisions(10),
		gridOn(false),
		axesOn(false)
	{};
	
	Void	Draw(Renderer &r, SLContext *);
	
	Int		divisions;
	Bool	gridOn;
	Bool	axesOn;
};

static Void DefineArrow(StrConst arrowName, GCLReal length, GCLReal width, GCLReal tipLength)
{	
	if (slObjectExists(arrowName)) return;

	scScenePtr	arrow;

	arrow = slBeginObject(arrowName);
	
	slBeginObject("shaft");
		slColour(cWhite);
		slTransform(Scale(width, length - tipLength, width) * Shift(0.0, 0.5, 0.0));
		slObject("cylinder");
	slEndObject();
	
	slBeginObject("tip");
		slTransform(Shift(0, length - tipLength, 0) * Scalef(tipLength));
		slObject("cone");
	slEndObject();

	slEndObject();

	slAddToLibrary(arrow);
}

static Void DefineAxes()
{
	scScenePtr axes;
	
	DefineArrow("gridArrow", 0.5, 0.02, 0.1);

	axes = slBeginObject("gridAxes");
	slColour(cGreen);
	slObject("gridArrow");
	slTransform(Rotation(vl_x, vl_halfPi));
	slColour(cBlue);
	slObject("gridArrow");
	slTransform(Rotation(vl_z, -vl_halfPi));
	slColour(cRed);
	slObject("gridArrow");
	slEndObject();

	slAddToLibrary(axes);
}

Void scFrame::Draw(Renderer &r, SLContext *)
{
	if (axesOn)
	{
		if (!slObjectExists("gridAxes"))
			DefineAxes();

		r.Draw(slGetLibrary()->Member("gridAxes"));
	}
		
	if (gridOn)
	{		
		Int		i, j;
		GCLReal	x, y, d;
		Point	p1, p2, p3, p4;

		d = 2.0 / divisions;
		
		r.C(Colour4(cGreen, 0.5));
		r.N(vl_y);
		p1 = p2 = p3 = p4 = vl_0;
		r.Begin(renLines);
		
		for (i = 0; i <= divisions; i++)
		{
			y = -1.0 + i * d;
			r.P(Point(-1.0, 0.0, y)).P(Point(1.0, 0.0, y));
			for (j = 0; j <= divisions; j++)
			{
				x = -1.0 + j * d;
				r.P(Point(x, 0.0, -1.0)).P(Point(x, 0.0, 1.0));
			}
		}
		r.End();

		r.C(Colour4(cCyan, 0.3));
		r.Begin(renPoly)
			.P(Point(-1.0, 0.0, -1.0))
			.P(Point(-1.0, 0.0,  1.0))
			.P(Point( 1.0, 0.0,  1.0))
			.P(Point( 1.0, 0.0, -1.0))
			.End();   
	}
}


// --- ScenePane class --------------------------------------------------------


ScenePane::ScenePane(Bool doubleBuf) :
	GLXPane(doubleBuf),
	itsScene(0),
	itsCamera(0),
	rot(vl_0), 
	trans(vl_0), 
	zoom(1.0), 
	wire(false),
	gridShown(false),
	axesShown(false)
{
	SetBgColour(cGrey50);
}

Void ScenePane::SetScene(scScenePtr scene)
{
	itsCamera = (scCamera *) scene->FindFirst(aCamera);

	if (!itsCamera)
		itsCamera = new scCamera;
	else
	{
		itsCamera->GetModelCentricParams();

		trans = itsCamera->trans;
		rot = itsCamera->rot;
	}
	
	itsScene = new scGroup;
	// add default colour
	itsScene->Add(new scColour(scPrimitive::sDefaultColour));
	// scene & camera
	itsScene->Add(itsCamera);
	itsScene->Add(scene);
	// add frame
	frame = new scFrame;
	frame->gridOn = gridShown;
	frame->axesOn = axesShown;
	itsScene->Add(frame);
	
	SetCameraParams();

	// make a list of mr models in the scene
	scene->ApplyAction(models);
	
	Redraw();

	if (doubleBuffered)
		Redraw();
}

Void ScenePane::SetCameraParams()
{
	itsCamera->SetAspect(GCLReal(width) / GCLReal(height));
	itsCamera->SetModelCentric(rot, trans);
}

Void ScenePane::Redraw()
{
	Clear().Draw(itsScene).Show();
}

const GCLReal kElevLimit = 0.495; 
	// To stop us reaching the singularity on the y axis.


Void ScenePane::ViewFrom(Point &position, Vector &normal)
{
// BROKEN: view stack
}

Void ScenePane::Restore()
{
// BROKEN
}

// --- Scene dragging ---------------------------------------------------------

Void ScenePane::TrackMouse(Int sx, Int sy, Int modifiers)
{
	Int	 		x, y;
	Coord		start((sx + 0.5) / width, 1.0 - (sy + 0.5) / height);
	Coord		c = start;
	Int			newModifiers;

	StartDrag(start, modifiers);
	GetMouse(x, y, newModifiers);

	while (newModifiers & wmButtons)
	{
		c = Coord((x + 0.5) / width, 1.0 - (y + 0.5) / height);
		Drag(c, modifiers);
		GetMouse(x, y, newModifiers);
	} 
	
	EndDrag(c, modifiers);
}

Void ScenePane::StartDrag(Coord c, Int mod)
{
	tRot = rot;
	tZoom = zoom;
	tTrans = trans;

	dragStart = c;
	modifiersStart = mod;

	baseErr = scMRModel::GetComplexity();

	if (mod & (wmButton1 | wmButton2))
		if (baseErr > 0.5)
			scMRModel::SetComplexity(0.1);
}

Void ScenePane::Drag(Coord c, Int)
{
	// standard behaviour: button1 = rotation, button 2 = translation,
	// button 3 = adjust complexity.

	offset = c - dragStart;
	if (modifiersStart & wmButton1)
	{
		if (modifiersStart & wmControl)
			SceneTransDrag();
		else
			SceneRotDrag();
	}
	else if (modifiersStart & wmButton2)
		SceneTransDrag();
	else if (modifiersStart & wmButton3)
		 ComplexityDrag();
}

Void ScenePane::EndDrag(Coord, Int)
{
	rot = tRot;
	trans = tTrans;
	zoom = tZoom;

	if (modifiersStart & (wmButton1 | wmButton2))
		if (baseErr != scMRModel::GetComplexity())
		{
			scMRModel::SetComplexity(baseErr);
			Redraw();
		}
}
	
Void ScenePane::ComplexityDrag()
{
	scMRModel::SetComplexity(Clip(baseErr + offset[1],
		(GCLReal) 0.0, (GCLReal) 1.0));

	Redraw();
}

Void ScenePane::SceneRotDrag()
{
    if (modifiersStart & wmShift)
	{
		tZoom = zoom - offset[1];
		itsCamera->SetScale(tZoom);
//		itsCamera->SetFOV(30.0 / tZoom);
	}
	else
	{
		tRot[1] = rot[1] + offset[0];	// rotation
		tRot[0] = rot[0] - offset[1];	// elevation
		if (tRot[0] < -kElevLimit)
			tRot[0] = -kElevLimit;
		else if (tRot[0] > kElevLimit)
			tRot[0] = kElevLimit;
	
		itsCamera->SetModelCentric(tRot, trans);
	}
			
	Clear().Draw(itsScene).Show();
}

Void ScenePane::SceneTransDrag()
{
	if (modifiersStart & wmShift)
		tTrans[2] = trans[2] - offset[1];
	else
	{
		tTrans[0] = trans[0] + offset[0];
		tTrans[1] = trans[1] + offset[1];
	}
	itsCamera->SetModelCentric(rot, tTrans);

	Clear().Draw(itsScene).Show();
}

Void ScenePane::HandleExpose()
{
	Redraw();
}

Void ScenePane::HandleKey(Char c)
{
	Int		i;
	
	switch (c)
	{
	case 'w':
		wire = !wire;
		if (wire)
			scPrimitive::sRenderStyle = renLineLoop;
		else
			scPrimitive::sRenderStyle = renPoly;
		Redraw();
		break;

	case 'G':
		SetGrid(!gridShown);
		Redraw();
		break;

	case 'A':
		SetAxes(!axesShown);
		Redraw();
		break;

	case 'S':
		for (i = 0; i < models.NumModels(); i++)
			models.Model(i)->SimplestModel();
		scMRModel::SetComplexity(0.0);
		Redraw();
		break;

	case 'C':
		for (i = 0; i < models.NumModels(); i++)
			models.Model(i)->MostComplexModel();
		scMRModel::SetComplexity(1.0);
		Redraw();
		break;

	case 'V':
		for (i = 0; i < models.NumModels(); i++)
			models.Model(i)->showMeta = !models.Model(i)->showMeta;
		Redraw();
		break;

	case 'F':
		for (i = 0; i < models.NumModels(); i++)
			models.Model(i)->showFaces = !models.Model(i)->showFaces;
		Redraw();
		break;

	case 'X':
		for (i = 0; i < models.NumModels(); i++)
			models.Model(i)->colourFaces = !models.Model(i)->colourFaces;
		Redraw();
		break;

	case 'N':
		for (i = 0; i < models.NumModels(); i++)
			models.Model(i)->CropClusters(models.Model(i)->currentClusters);
		Redraw();
		break;

	case '[':
		for (i = 0; i < models.NumModels(); i++)
			models.Model(i)->AdaptClusters(models.Model(i)->currentClusters - 1);
		Redraw();
		break;
	case ']':
		for (i = 0; i < models.NumModels(); i++)
			models.Model(i)->AdaptClusters(models.Model(i)->currentClusters + 1);
		Redraw();
		break;
		
	case '0':
		SetHeadlight(false);
		Redraw();
		break;
			
	case '1':
		SetHeadlight(true);
		Redraw();
		break;

	case '!':
		SetDoubleSided(true);
		Redraw();
		break;
			
	case ')':
		SetDoubleSided(false);
		Redraw();
		break;

	default:
		XEventPane::HandleKey(c);
	}
}

Void ScenePane::SetGrid(Bool on)
{
	gridShown = on;
	frame->gridOn = on;
}

Void ScenePane::SetAxes(Bool on)
{
	axesShown = on;
	frame->axesOn = on;
}

// --- ScenePane2D methods ----------------------------------------------------

Void ScenePane2D::SetScene(scScenePtr scene)
{
	ScenePane::SetScene(scene);
	itsCamera->SetFOV(RadsToDegs(atan(0.25)));
}

Void ScenePane2D::Drag(Coord c, Int)
{
	offset = c - dragStart;

	if (modifiersStart & wmShift)
	{
		tTrans[2] = trans[2] + offset[1];
	}
	else
	{
		tTrans[0] = trans[0] + offset[0] / zoom;
		tTrans[1] = trans[1] + offset[1] / zoom;
	}

	itsCamera->SetModelCentric(rot, tTrans);

	Clear().Draw(itsScene).Show();
}

#endif
