/*
	File:		FCElem.h

	Function:	Elements that implement face cluster radiosity

	Author:		Andrew Willmott

	Copyright:	(c) 1998-2000 Andrew Willmott
*/

#ifndef __VecRadElem__
#define __VecRadElem__

#include "HRElem.h"
#include "RadMesh.h"
#include "RT_Prim.h"

#include "gcl/MRModel.h"
#include "cl/Flags.h"

typedef ClrMat IrradVec;
class FCElem;
class FCShadeInfo;
class FCSIRec;

struct MRInfo
//
//	info about how model is embedded in the current scene. needed 'cause
//
//	(1) assume we don't have access to scene data from RadMethod
//
//	(2) same model might be positioned in several places in the scene
//
//	(3) we're hoping to only access small parts of the MRVertex info, so
//	don't want to be making local, transformed copies of all of them all
//	the time.
//
{
	~MRInfo();

	Transform	trans;		///< transform at point of attack
	VecTrans	orient;		///< orientation transform, for mapping vectors
	GCLReal		areaScale;	///< (uniform) scale factor of the same for areas
	Colour		refl;		///< reflectance for this version of the model
	Colour		emit;		///< emittance for this version of the model
	ColourList	colours;	///< vertex colours calculated by rad. sim.
	IndexList	weights;	///< weights for face-cluster vertex colour calcs.
	MRModel		*model;		///< original model;
	RT_Object	rtObj;		///< corresponding ray-trace object.
	RadProps	props;		///< for any leaf subdivision

#ifdef RAD_VIS
	HRElem				*FindElem(Int fci);	///< Find element from a leaf index
	PtrArray<HRElem>	elemPtrs;
#endif	
};

class FCElem : public HRElem
{
public:
	FCElem(Int fci, MRInfo *myInfo);

	Void		Print(ostream &s);
	Void		DrawNodeElem(Renderer &r);
	Void		Reset();
	
    Bool		IsLeaf()		
				{ return(child[0] == 0); }
	Void		ApplyToChildren(Void (HRElem::*method)(Void*), Void *ref = 0);
	   
	GCLReal		Error();
	Void		Add();
	Void		Push();
	Void		Pull();
	Void		ClearB();
	Void		ClearR();
	Void		CalcLeafRadiosity();

	Void		MakeChildLinks(HRElem *other, HRLink *link,
							   Int which, Int levels);

	// HRElem stuff
	Void		EltUpdateBounds(Point &min, Point &max);
	GCLReal		EltArea()	
				{ return(totArea); };
	Point		EltCentre() 
				{ return(centre); };
	Colour		EltE()		
				{ return(info->emit); };
	Colour		EltRho();
	Void		EltSetVisPoints(HRElem *to, Point pts[]);
	Void		EltSampleTransport(Int numSamples, Point p[], Vector n[],
						Matd &coeffs);
	GCLReal		EltCalcTransport(HRElem *from, Matd &coeffs);

	Colour 		EltBA();

	// vector radiosity/cluster
	GCLReal		EltProjArea(const Vector &v)
				{ return(Max(0.0, dot(areaNormal, v))); };
	Void		AddIrradiance(const Colour &E, const Vector &m);
	Void		AddChanIrradiance(const Colour &E, Int chan);
	Colour		GetPower(const Vector &m);
	Void		EltGetSamples(Int numSamples, Point p[]);

	GCLReal 	EltMaxProjArea(const Vector &m);

	Colour		*B_Coeffs();
	Colour		*R_Coeffs() 
				{ _Error("argh"); return(0); };
	Int			NumCoeffs()
				{ return(1); };

	// internal

	Point		ToWorld(const Point &p)
				{ return(xform(info->trans, p * axes)); };

	Void		ColourFaces(MRModel *model, Int fci);
	Void		ColourFacesBest(MRModel *model, Int fci, const FCShadeInfo &si);

	// asd

	Void		Subdivide();

 	Void		DistributeColours();  
 	Void		DistributeColoursBest(ShadeInfo &shadeInfo);
	Void		CalcShadeInfo(FCShadeInfo &si, const HRLinkList &links);
	Void		CalcShadeInfo(FCShadeInfo &si, const FCSIRec &parIrr);

	Void		DrawLeafElem(Renderer &r);

	Void		DebugInfo();


	// Fields

	// associated cluster info
	MRInfo		*info;
	FaceCluster	&fc;
	Int			fci;

	Point		centre;

	Vector		areaNormal;		// A * n
	Float		sideArea[6];	// for proj. area upper bound
	Float		projArea;		// len(areaNormal)
	Float		totArea;		// total (unprojected) area
	Float		maxVisArea;		// max externally visible area
	VecTrans	axes;
	
	HRElem		*child[2];
	Colour		B;
	IrradVec    R;
	static Colour lastB;
};

class ModelsExtractor : public MRModelsFinder
// Extracts MR models from scene, and builds MRInfo for each one.
{
public:
	Void				Start();
	Void				Primitive(scPrimitive *sp);

	PtrArray<MRInfo>	info;
};

#endif

