/*
	File:			Flatlets.cc

	Function:		See header file

	Author(s):		Andrew Willmott

	Copyright:		(c) 1996-2000, Andrew Willmott

*/

#include "Flatlets.h"
#include "WaveCoeffs.h"
#include "WaveSupport.h"
#ifdef RAD_VIS
#include "RadMethod.h"
#endif

#define DECLARE_T 	GCLMat	T(To()->Coeffs(), From()->Coeffs(), transport)
#define F2Child(x) ((F2Elem *) child[x])


// --- F2Link methods ---------------------------------------------------------


Bool F2Link::CalcTransport()
//	Calculate F2 transport coefficients.
{
	DECLARE_T;
	Int			i;
	GCLReal		*visPtr = 0;
	
	if (gRadControl->visInQuad)
	{
		// calculate the visibility as we go
		visibility = 0.0;
		visPtr = &visibility;
	}
	
	From()->SubToSubFormFactor(2, To(), 2, T, visPtr);
	factor = transport[0];
	for (i = 1; i < 16; i++)
		if (factor < transport[i])
			factor = transport[i];
	return(factor > 0.0);
}

Bool F2Link::CalcVisibility(HRLink *parent, Bool forceVisReuse)
{
	if (parent && (forceVisReuse || !NeedVisibility(parent)))
		visibility = parent->visibility;
	else
		visibility = From()->Visibility(To());
	
	DECLARE_T;

	T *= visibility;

	return(true);
}

Void F2Link::Gather()
{
	DECLARE_T;
	// R += T * B
	ClrVecMultAcc(T, From()->B, To()->R);
}

GCLReal F2Link::Error()
{
	return(factor);
}

GCLReal F2Link::BFAError()
{
	return(To()->area * factor * dot(kRadRGBToLum, To()->Reflectance() * 
		ClrVecAverage(From()->B, From()->Coeffs())));
}

Void F2Link::Print(ostream &s)
{
	DECLARE_T;
	HRLink::Print(s);
	s << " transport = " << T << " vis = " << visibility; 
}

Void F2Link::DrawLink(Renderer &r, GCLReal left, GCLReal top, 
		GCLReal right, GCLReal bottom, GCLReal weight)
{
	DECLARE_T;
	
	DrawWaveLink(r, left, top, right, bottom, To()->Reflectance() * weight, T);
}


// --- F2Elem methods ---------------------------------------------------------


Colour F2Elem::lastB[4];

Void F2Elem::Add()
{	
	for (Int i = 0; i < Coeffs(); i++)
		B[i] += R[i];
}
										
Void F2Elem::Push()
{
	if (IsTri())
	{
		//	c[i].B = F2_TRI_P[i] * B;
		ClrVecMult(F2_TRI_P0, B, F2Child(1)->B);
		ClrVecMult(F2_TRI_P1, B, F2Child(2)->B);
		ClrVecMult(F2_TRI_P2, B, F2Child(0)->B);
		ClrVecMult(F2_TRI_P3, B, F2Child(3)->B);
		// total: 3(64*, 48+)
	}
	else
	{	
		/*
			Conceptually, this is just trans(T) * B * T: T interpolates rows, 
			and trans(T) interpolates columns. We break up the multiplication
			because our destination matrix is not stored contigously; each
			quadrant is stored in a separate child.
		*/
		Colour t[4];	// temp

		ClrVecRightTensorMult(F2_LINE_P0, B, t);	
		ClrVecLeftTensorMult(F2_LINE_P0, t, F2Child(1)->B);
		ClrVecLeftTensorMult(F2_LINE_P1, t, F2Child(0)->B);
		ClrVecRightTensorMult(F2_LINE_P1, B, t);
		ClrVecLeftTensorMult(F2_LINE_P0, t, F2Child(2)->B);
		ClrVecLeftTensorMult(F2_LINE_P1, t, F2Child(3)->B);
	}
}

Void F2Elem::Pull()
{
	if (IsTri())
	{
		// B = sum F2_TRI_L[i] * c[i].B
		ClrVecMult   (F2_TRI_L0, F2Child(1)->B, B);
		ClrVecMultAcc(F2_TRI_L1, F2Child(2)->B, B);
		ClrVecMultAcc(F2_TRI_L2, F2Child(0)->B, B);
		ClrVecMultAcc(F2_TRI_L3, F2Child(3)->B, B);
	}
	else
	{// budget: 48/36
		Colour t[4];
		
		ClrVecRightTensorMult(F2_LINE_L0, F2Child(1)->B, t);
		ClrVecRightTensorMultAcc(F2_LINE_L1, F2Child(2)->B, t);
		ClrVecLeftTensorMult(F2_LINE_L0, t, B);
		
		ClrVecRightTensorMult(F2_LINE_L0, F2Child(0)->B, t);
		ClrVecRightTensorMultAcc(F2_LINE_L1, F2Child(3)->B, t);
		ClrVecLeftTensorMultAcc(F2_LINE_L1, t, B);
	}
}

Void F2Elem::ClearB()
{
	for (Int i = 0; i < Coeffs(); i++)
	{
		lastB[i] = B[i];
		B[i] = vl_0;
	}
}

Void F2Elem::CalcLeafRadiosity()
{
	for (Int i = 0; i < Coeffs(); i++)
		B[i] = B[i] * Reflectance() + Emittance();
}

GCLReal F2Elem::Error()
{
	return(ClrVecCalcError(B, lastB, Coeffs()));
}

Void F2Elem::ClearR()
{
	for (Int i = 0; i < Coeffs(); i++)
		R[i] = vl_0;
}


// --- The numerical code for F2s ---------------------------------------------


Colour F2Elem::SampleLeaf(Coord c)
{
	if (gRadControl->gouraud)
	{
		Colour mB[4];

		// Transform into equiv. multiwavelet basis...
		if (IsQuad())
		{
			ClrVecTensorMult(MCC2_LINE_R, B, mB);
			return(M2QuadSample(c, mB));
		}
		else
		{
			ClrVecMult(MCC2_TRI_R, B, mB);
			return(M2TriSample(c, mB));
		}
	}
	else
	{
		Int i = 0;	
		if (IsQuad())
		{
			if (c[0] >= 0.5)
				i++;
			if (c[1] >= 0.5)
				i += 2;
		}
		else
		{
			if (c[0] >= 0.5)
				i = 1;
			else if (c[1] >= 0.5)
				i = 2;
			else if (c[0] + c[1] >= 0.5)
				i = 3;	
		}

		Assert(i < Coeffs(), "(SampleLeaf) Bad i");
		return(B[i]);
	}
}

Void F2Elem::DrawLeaf(Renderer &r)
{
	if (gRadControl->wire)
		RadElem::Draw(r);
	else
	{
		Int		i, j;
		Colour	*c = B;
		
		Vector dx = (Vertex(2) - Vertex(1)) / 2.0;
		Vector dy = (Vertex(0) - Vertex(1)) / 2.0;
		Point pt;
		
		if (gRadControl->funcView)
		{
			r.Begin(renLineLoop).C(cWhite);
			SendPoints(r);
			r.End();
		}

		if (highlight)
		{
			DrawHighlight(r);
			if (!gRadControl->funcView)
				return;
		}

		if (gRadControl->gouraud)
			DrawSampledLeaf(r, 1);
		else if (gRadControl->funcView)
			for (i = 0; i < 2; i++)
				for (j = 0; j < 2; j++)
				{
					r.Begin(renPoly);
					if (IsQuad())
					{
						pt = Vertex(1) + dx * j + dy * i;
						Project(r, *c, pt);
						Project(r, *c, pt + dx);
						Project(r, *c, pt + dx + dy);
						Project(r, *c, pt + dy);
					}
					else if (i + j >= 2)
					{
						pt = Vertex(1) + dx * (2 - i) + dy * (2 - j);
						Project(r, *c, pt);
						Project(r, *c, pt - dx);
						Project(r, *c, pt - dy);
					}
					else
					{
						pt = Vertex(1) + dx * j + dy * i;
						Project(r, *c, pt);
						Project(r, *c, pt + dx);
						Project(r, *c, pt + dy);
					}
					c++;
					r.End();
				}
		else						
			for (i = 0; i < 2; i++)
				for (j = 0; j < 2; j++)
				{
					r.SetColour(*c).Begin(renPoly);
					
					if (IsQuad())
					{
						pt = Vertex(1) + dx * j + dy * i;
						r.P(pt).P(pt + dx).P(pt + dx + dy).P(pt + dy);
					}
					else if (i + j >= 2)
					{
						pt = Vertex(0) + dx * (2 - i) + dy * (2 - j);
						r.P(pt).P(pt - dx).P(pt - dy);
					}
					else
					{
						pt = Vertex(1) + dx * j + dy * i;
						r.P(pt).P(pt + dx).P(pt + dy);
					}
					c++;
					r.End();
				}
	}
}

Void F2Elem::Print(ostream &s)
{
	s << " + f2 ";
	PrintRec(s);
}

Void F2Elem::PrintSelf(ostream &s)
{
	HRMeshElem::PrintSelf(s);
	for (Int i = 0; i < Coeffs(); i++)
		s << B[i] << ' ';
}

Void F2Elem::ParseSelf(istream &s)
{
	HRMeshElem::ParseSelf(s);
	for (Int i = 0; i < Coeffs(); i++)
		s >> B[i];
}

Void F2Elem::SetParent(HierElem &itsParent)
{
	HierElem::SetParent(itsParent);

//	if (itsParent.child[3] == this)
//		((F2Elem &) itsParent).Push();
}

Void F2Elem::DistributeColours()
{
	colour = cRed;
	if (HasChildren())
	{
		F2Child(0)->DistributeColours();
		F2Child(1)->DistributeColours();
		F2Child(2)->DistributeColours();
		F2Child(3)->DistributeColours();
	}
}

// --- generic interface ------------------------------------------------------

					// find coeffs s.t. dot(from->B, coeffs) = irradiance
					// at p[], n[]. If n = 0, calculate isotropic point
					// irradiance.
Void F2Elem::EltSampleTransport(
		Int		numSamples,
		Point	p[],
		Vector	n[],
		Matd	&coeffs
	)
// find coeffs s.t. dot(from->B, coeffs[i]) = irradiance
// at p[i], n[i].
{
#ifdef UNFINISHED
	Int		i;

	for (i = 0; i < numSamples; i++)
		coeffs[i][0] = ApproxPatchFactor(p[i], n[i]);

	Coord	*tuv, *fuv;
	Int		i, j, k, fn, tn;

	if (IsQuad())
	{ fx = F2_LINE_x.Ref(); fy = F2_LINE_y.Ref(); fn = F2_LINE_x.Elts(); }
	else
	{ fx = MT2_x.Ref(); fy = MT2_y.Ref(); fn = MT2_x.Elts(); }
	
	SampleKernel(fromU, To(), tuv, tn, K.Ref(), visPtr);
#endif
}

GCLReal F2Elem::EltCalcTransport(HRElem *from, Matd &coeffs)
{
#ifdef UNFINISHED
	if (from == this)
	{
		// no self-reflection for flat poly...
		coeffs = vl_0;
		return(0.0);
	}

	from->EltSampleTransport(1, &centre, &normal, coeffs);
	DBG_COUT << "haar: transport coeffs = " << coeffs << endl;
	return(coeffs[0][0]);
#endif
}

Void F2Elem::DistributeR(const Colour &srcR, const Vector &r)
{
	GCLReal		temp;
	Int			i;
	
	temp = dot(r, normal);
	// coefficient R[0] is the constant function...
	if (temp > 0.0)
		for (i = 0; i < Coeffs(); i++)
			R[i] += temp * srcR;
}

// --- F3 Flatlet methods -----------------------------------------------------

#define F3Child(x) ((F3Elem *) child[x])

// --- F3Link methods ---------------------------------------------------------


Bool F3Link::CalcTransport()
//	Calculate F3 transport coefficients.
{
	DECLARE_T;
	Int			i;
	GCLReal		*visPtr = 0;
	
	if (gRadControl->visInQuad)
	{
		visibility = 0.0;
		visPtr = &visibility;
	}

	From()->SubToSubFormFactor(3, To(), 3, T, visPtr);
	factor = transport[0];
	for (i = 1; i < 81; i++)
		if (factor < transport[i])
			factor = transport[i];
	return(factor > 0.0);	
}

Bool F3Link::CalcVisibility(HRLink *parent, Bool forceVisReuse)
{
	if (parent && (forceVisReuse || !NeedVisibility(parent)))
		visibility = parent->visibility;
	else
		visibility = From()->Visibility(To());
	
	DECLARE_T;

	T *= visibility;

	return(true);
}

Void F3Link::Gather()
{
	DECLARE_T;
	// R += T * B
	ClrVecMultAcc(T, From()->B, To()->R);
}

GCLReal F3Link::Error()
{
	return(factor);
}

GCLReal F3Link::BFAError()
{
	return(To()->area * factor * dot(kRadRGBToLum, To()->Reflectance() * 
		ClrVecAverage(From()->B, From()->Coeffs())));
}

Void F3Link::Print(ostream &s)
{
	DECLARE_T;
	HRLink::Print(s);
	s << " transport = " << T << " vis = " << visibility;
}

Void F3Link::DrawLink(Renderer &r, GCLReal left, GCLReal top,
		GCLReal right, GCLReal bottom, GCLReal weight)
{
	DECLARE_T;
	
	DrawWaveLink(r, left, top, right, bottom, To()->Reflectance() * weight, T);
}


// --- F3Elem methods ---------------------------------------------------------


Colour F3Elem::lastB[9];

Void F3Elem::Add()
{	
	for (Int i = 0; i < Coeffs(); i++)
	{
		// 3rd-order bases can actually go negative, esp. round corners
		// if we don't nip this in the bud by clamping here, solution
		// will spiral out of control.
//		ClipColourZero(R[i]);

		if (R[i][0] < 0.0) R[i][0] = 0.0;
		if (R[i][1] < 0.0) R[i][1] = 0.0;
		if (R[i][2] < 0.0) R[i][2] = 0.0;

		B[i] += R[i];
		Assert(B[i][0] >= 0.0 && B[i][1] >= 0 && B[i][2] >=0, "-ve coeff!");
	}
}

Void F3Elem::Push()
{
	if (IsTri())
	{
		//	c[i].B = F3_TRI_P[i] * B;
		ClrVecMult(F3_TRI_P0, B, F3Child(1)->B);
		ClrVecMult(F3_TRI_P1, B, F3Child(2)->B);
		ClrVecMult(F3_TRI_P2, B, F3Child(0)->B);
		ClrVecMult(F3_TRI_P3, B, F3Child(3)->B);
		// total: 3(64*, 48+)
	}
	else
	{	
		/*
			Conceptually, this is just trans(T) * B * T: T interpolates rows, 
			and trans(T) interpolates columns. We break up the multiplication
			because our destination matrix is not stored contigously; each
			quadrant is stored in a separate child.
		*/
		Colour t[9];

		ClrVecRightTensorMult(F3_LINE_P0, B, t);
		ClrVecLeftTensorMult(F3_LINE_P0, t, F3Child(1)->B);	
		ClrVecLeftTensorMult(F3_LINE_P1, t, F3Child(0)->B);
		ClrVecRightTensorMult(F3_LINE_P1, B, t);
		ClrVecLeftTensorMult(F3_LINE_P0, t, F3Child(2)->B);
		ClrVecLeftTensorMult(F3_LINE_P1, t, F3Child(3)->B);
	}
}

Void F3Elem::Pull()
{
	if (IsTri())
	{
		ClrVecMult   (F3_TRI_L0, F3Child(1)->B, B);
		ClrVecMultAcc(F3_TRI_L1, F3Child(2)->B, B);
		ClrVecMultAcc(F3_TRI_L2, F3Child(0)->B, B);
		ClrVecMultAcc(F3_TRI_L3, F3Child(3)->B, B);
	}
	else
	{
		Colour t[9];
		
		ClrVecRightTensorMult(F3_LINE_L0, F3Child(1)->B, t);	
		ClrVecRightTensorMultAcc(F3_LINE_L1, F3Child(2)->B, t);	
		ClrVecLeftTensorMult(F3_LINE_L0, t, B);
		
		ClrVecRightTensorMult(F3_LINE_L0, F3Child(0)->B, t);	
		ClrVecRightTensorMultAcc(F3_LINE_L1, F3Child(3)->B, t);	
		ClrVecLeftTensorMultAcc(F3_LINE_L1, t, B);
	}
}

Void F3Elem::ClearB()
{
	for (Int i = 0; i < Coeffs(); i++)
	{
		lastB[i] = B[i];
		B[i] = vl_0;
	}
}

Void F3Elem::CalcLeafRadiosity()
{
	for (Int i = 0; i < Coeffs(); i++)
	{
//		FindMaxCmpts(B[i], vl_0, B[i]);
		B[i] = B[i] * Reflectance() + Emittance();
	}
}

GCLReal F3Elem::Error()
{
	return(ClrVecCalcError(B, lastB, Coeffs()));
}

Void F3Elem::ClearR()
{
	for (Int i = 0; i < Coeffs(); i++)
		R[i] = vl_0;
}


// --- The numerical code for F2s ---------------------------------------------


Colour F3Elem::SampleLeaf(Coord c)
{
	if (gRadControl->gouraud)
	{
		Colour	mB[9];

		// Transform to multiwavelet
		if (IsQuad())
		{
			ClrVecTensorMult(MCC3_LINE_R, B, mB);
			return(M3QuadSample(c, mB));
		}
		else
		{
			ClrVecMult(MCC3_TRI_R, B, mB);		
			return(M3TriSample(c, mB));
		}
	}
	else
	{
		Int m = 0, ci[2], n = 3;
		
		c *= (GCLReal) n;
		ci[0] = (Int) c[0]; if (ci[0] == n) ci[0]--;
		ci[1] = (Int) c[1]; if (ci[1] == n) ci[1]--;

		if (IsTri() && (c[0] + c[1] >= ci[0] + ci[1]))
			m = (n + 1) * (n - 1) - n * ci[0] - ci[0];
		else
			m = n * ci[1] + ci[0];

		Assert(m < Coeffs(), "(SampleLeaf) bad m");
		return(B[m]);
	}
}

Void F3Elem::DrawLeaf(Renderer &r)
{
	if (gRadControl->wire)
		RadElem::Draw(r);
	else
	{
		Int		i, j;
		Colour	*c = B;
		Vector	dx = (Vertex(2) - Vertex(1)) / 3.0;
		Vector	dy = (Vertex(0) - Vertex(1)) / 3.0;
		Point	pt;
		
		if (gRadControl->funcView)
		{
			r.Begin(renLineLoop).C(cWhite);
			SendPoints(r);
			r.End();
		}
		
		if (highlight)
		{
			DrawHighlight(r);
			if (!gRadControl->funcView)
				return;
		}
		
		if (gRadControl->gouraud)
			DrawSampledLeaf(r, 4);
		else if (gRadControl->funcView)
			for (i = 0; i < 3; i++)
				for (j = 0; j < 3; j++)
				{
					r.Begin(renPoly);
					if (IsQuad())
					{
						pt = Vertex(1) + dx * j + dy * i;
						Project(r, *c, pt);
						Project(r, *c, pt + dx);
						Project(r, *c, pt + dx + dy);
						Project(r, *c, pt + dy);
					}
					else if (i + j >= 3)
					{
						pt = Vertex(1) + dx * (3 - i) + dy * (3 - j);
						Project(r, *c, pt);
						Project(r, *c, pt - dx);
						Project(r, *c, pt - dy);
					}
					else
					{
						pt = Vertex(1) + dx * j + dy * i;
						Project(r, *c, pt);
						Project(r, *c, pt + dx);
						Project(r, *c, pt + dy);
					}
					c++;
					r.End();
				}
		else
			for (i = 0; i < 3; i++)
				for (j = 0; j < 3; j++)
				{
					r.SetColour(*c++).Begin(renPoly);
					if (IsQuad())
					{
						pt = Vertex(1) + dx * j + dy * i;
						r.P(pt).P(pt + dx).P(pt + dx + dy).P(pt + dy);
					}
					else if (i + j >= 3)
					{
						pt = Vertex(1) + dx * (3 - i) + dy * (3 - j);
						r.P(pt).P(pt + dx).P(pt + dy);
					}
					else
					{
						pt = Vertex(1) + dx * j + dy * i;
						r.P(pt).P(pt - dx).P(pt - dy);
					}
					r.End();
				}	
	}
}

Void F3Elem::Print(ostream &s)
{
	s << " + f3 ";
	PrintRec(s);
}

Void F3Elem::PrintSelf(ostream &s)
{
	HRMeshElem::PrintSelf(s);
	for (Int i = 0; i < Coeffs(); i++)
		s << B[i] << ' ';
}

Void F3Elem::ParseSelf(istream &s)
{
	HRMeshElem::ParseSelf(s);
	for (Int i = 0; i < Coeffs(); i++)
		s >> B[i];
}

Void F3Elem::SetParent(HierElem &itsParent)
{
	HierElem::SetParent(itsParent);
//	if (itsParent.child[3] == this)
//		((F3Elem &) itsParent).Push();
}

Void F3Elem::DistributeColours()
{
	colour = cRed;
	if (HasChildren())
	{
		F3Child(0)->DistributeColours();
		F3Child(1)->DistributeColours();
		F3Child(2)->DistributeColours();
		F3Child(3)->DistributeColours();
	}
}

// --- generic interface ------------------------------------------------------

					// find coeffs s.t. dot(from->B, coeffs) = irradiance
					// at p[], n[]. If n = 0, calculate isotropic point
					// irradiance.
Void F3Elem::EltSampleTransport(
		Int		numSamples,
		Point	p[],
		Vector	n[],
		Matd	&coeffs
	)
// find coeffs s.t. dot(from->B, coeffs[i]) = irradiance
// at p[i], n[i].
{
#ifdef UNFINISHED
	Int		i;

	for (i = 0; i < numSamples; i++)
		coeffs[i][0] = ApproxPatchFactor(p[i], n[i]);

	Coord	*tuv, *fuv;
	Int		i, j, k, fn, tn;

	if (IsQuad())
	{ fx = F3_LINE_x.Ref(); fy = F3_LINE_y.Ref(); fn = F3_LINE_x.Elts(); }
	else
	{ fx = MT2_x.Ref(); fy = MT2_y.Ref(); fn = MT2_x.Elts(); }
	
	SampleKernel(fromU, To(), tuv, tn, K.Ref(), visPtr);
#endif
}

GCLReal F3Elem::EltCalcTransport(HRElem *from, Matd &coeffs)
{
#ifdef UNFINISHED
	if (from == this)
	{
		// no self-reflection for flat poly...
		coeffs = vl_0;
		return(0.0);
	}

	from->EltSampleTransport(1, &centre, &normal, coeffs);
	DBG_COUT << "haar: transport coeffs = " << coeffs << endl;
	return(coeffs[0][0]);
#endif
}

Void F3Elem::DistributeR(const Colour &srcR, const Vector &r)
{
	GCLReal		temp;
	Int			i;
	
	temp = dot(r, normal);
	// coefficient R[0] is the constant function...
	if (temp > 0.0)
		for (i = 0; i < Coeffs(); i++)
			R[i] += temp * srcR;
}
