/*
	File:			HRLink.h

	Function:		Defines a link type for use with hierarchical/
					wavelet radiosity methods.
					
	Author(s):		Andrew Willmott

	Copyright:		(c) 1997-2000, Andrew Willmott
 */


#ifndef __HRLink__
#define __HRLink__

#include "cl/List.h"
#include "gcl/Renderer.h"


// --- Link between two hierarchical elements ---------------------------------


class HRElem;

enum RefChoice { kSubNone, kSubFrom, kSubTo, kSubBoth };

class HRLink : public Node
{
public:
	HRLink();

	virtual HRLink		*New() = 0;
						///< Create a new link of the same type.
	virtual Void		FreeLink(HRLink *link);
						///< Free up link

	Bool				MakeLink(
							HRElem	*fromElem,
							HRElem	*toElem,
							HRLink	*parentLink = 0,
							Bool	forceVisReuse = false
						);
						///< Setup a link.	
	virtual Bool	   	CalcTransport() = 0;
						/**< Calculate the transport coefficients. 
							 Returns true if the transport is non-zero. */
	virtual Bool	   	CalcVisibility(HRLink *parentLink, Bool forceVisReuse);
						/**< Calculate the fractional visibility, and
							 multiply into transport.  Return true if
							 the link isn't totally occluded. */
	virtual Void		Gather() = 0;
						/**< Gather radiosity from from->B and accumulate
							 in to->R */
	
	Bool				NeedVisibility(HRLink *parentLink);
						///< Return true if we need a new visibility estimate.
	
	virtual RefChoice	RefineOracle();
						///< Decide if and how the link needs to be refined.

	virtual GCLReal	  	Error() = 0;
	virtual GCLReal		BFAError() = 0;
	
	HRLink			   	*CreateSubLink(HRElem *from, HRElem *to);
						/**< Create a new link between from and to,
							 reusing the visibility calcs of the old link
							 if possible. */

	virtual Void	   	Print(ostream &s);
	virtual Void	   	DrawLink(Renderer &r, GCLReal left,
								GCLReal top, GCLReal right,
								 GCLReal bottom, GCLReal weight);
	virtual GCLReal		Strength() = 0;
	virtual Void		DebugInfo();

	// Fields
	HRElem			   	*from, *to;
	GCLReal			   	visibility;
	Flags8				flags;
};

ostream &operator << (ostream &s, HRLink &rl);

typedef List			HRLinkList;
typedef Iter<HRLink>	HRLinkIter;

ostream &operator << (ostream &s, HRLink &l);

inline Vector safe_norm(const Vector &v)	
{
	GCLReal		r;
	Vector		result;

	r = len(v);
	result = v;
	if (r > 0.0)
		result /= r;
		
	return(result);
}

#endif
