/*
	File:			HRMesh.cc

	Function:		See header file

	Author(s):		Andrew Willmott

	Copyright:		(c) 1997-2000, Andrew Willmott

*/

#include "HRMesh.h"
#include "RadMethod.h"
#include "Samples.h"
#include "gcl/Draw.h"

#ifdef RAD_VIS
#include "gcl/Forms.h"
#include "gcl/ScenePane.h"
#endif

#define MAX_ELT_SAMPLES 32


// --- HRMeshElem methods -----------------------------------------------------

HRMeshElem::HRMeshElem() : HRElem(), HierElem()
{
	flags.Set(hrPatch);
}

Void HRMeshElem::Reset()
{
	HRElem::Reset();
	HierElem::FreeChildren();
	flags.UnSet(hrIntNode);
}

Void HRMeshElem::ApplyToChildren(Void (HRElem::*method)(Void *), Void *ref)
{
	if (!IsLeaf())
		for (Int i = 0; i < 4; i++)
			(((HRMeshElem*)child[i])->*method)(ref);
}

// ----------------------------------------------------------------------------

Void HRMeshElem::Print(ostream &s)
{
	HRLinkIter i;
	
	PrintID(s);
	s << " = ";
	HierElem::PrintSelf(s);
	s << ' ';
	for (i.Begin(links); !i.AtEnd(); i.Inc())
		s << i.Data();
		
	s << endl;
}

Void HRMeshElem::CreatePatches(PatchList &patches)
{
	patches.Append(this);
}


// --- kernel sampling --------------------------------------------------------


Void HRMeshElem::MakeChildLinks(HRElem *other, HRLink *link, 
								Int which, Int levels)
{
	Int i;
	
	if (IsLeaf())
	{
		if (!HasChildren())
			Subdivide();
		flags.Set(hrIntNode);
		Push();
	}

	if (which == kSubTo)
		for (i = 0; i < 4; i++)
			Child(i)->RefineLink(
				link->CreateSubLink(other, Child(i)),
				levels);		
	else if (which == kSubFrom)
		for (i = 0; i < 4; i++)
			other->RefineLink(
				link->CreateSubLink(Child(i), other),
				levels);
	else if (which == kSubBoth)
		for (i = 0; i < 4; i++)
			other->MakeChildLinks(Child(i), link, kSubFrom, levels);
	else
		_Error("split type not implemented");
}

Void HRMeshElem::SampleKernel(
			const GCLMat		&fromU,		// source patch samples			
			HRMeshElem 			*to, 		// dest. patch
			const GCLMat		&toU,		// destination samples
			GCLMat				&samples,	// results of sampling the kernel.
			GCLReal 			*visPtr		// if visPtr is set, accumulate
											// visibility into it
		)
//	x, y E [0, 1], origin at patch's lower left corner (vertex 1).
{
	GCLReal	result, vf = 1.0 / (fromU.Rows() * toU.Rows()), xs, ys;
	Vector	temp, delta;
	Point	srcPt[MAX_ELT_SAMPLES], srcPtE, srcPtV;
	Vector	dstPt[MAX_ELT_SAMPLES], dstPtE, dstPtV;
	Vector	dx2 = to->Vertex(2) - to->Vertex(1);	
	Vector	dy2 = to->Vertex(0) - to->Vertex(1);
	Vector	dx1 = 	  Vertex(2) -     Vertex(1);	
	Vector 	dy1 = 	  Vertex(0) -     Vertex(1);	
	Int		i, j;

#ifdef RAD_VIS
	if ((gRadControl->showRays && visPtr) || gRadControl->showLinks)
	{
		RM_DISPLAY_START;
		gRadControl->radObject->display->Begin(renLines);
	}
#endif

	samples.SetSize(toU.Rows(), fromU.Rows());
	
	// calculate sample positions on source patch (that's us).
	srcPtE = Normal(); 
	srcPtE *= kRaySurfEps;
	for (j = 0; j < fromU.Rows(); j++)
	{
		srcPt[j] = Vertex(1);
		srcPt[j] += fromU[j][0] * dx1;
		srcPt[j] += fromU[j][1] * dy1;
	}
		
	// ditto for destination
	dstPtE = to->Normal(); 
	dstPtE *= kRaySurfEps;
	for (i = 0; i < toU.Rows(); i++)
	{
		dstPt[i] = to->Vertex(1);
		dstPt[i] += toU[i][0] * dx2;
		dstPt[i] += toU[i][1] * dy2;
	}

	for (i = 0; i < toU.Rows(); i++)
	{
		for (j = 0; j < fromU.Rows(); j++)
		{
#ifdef RAD_VIS
			if (gRadControl->showLinks)
				gRadControl->radObject->display->C(cYellow).P(dstPt[i]).P(srcPt[j]);
#endif
			
			delta = dstPt[i];
			delta -= srcPt[j];
			temp = safe_norm(delta);
			result = dot(Normal(), temp);
			
			if (result <= 0.0)
				result = 0.0;
			else 
			{
				result *= -dot(to->Normal(), temp);
				if (result <= 0.0)
					result = 0.0;
				else
				{
					result *= area / (sqrlen(delta) * vl_pi);
					
					if (visPtr)
					{
						Bool hit;
						
						// Sample visibility too...
						srcPtV = srcPt[j] + srcPtE;
						dstPtV = dstPt[i] + dstPtE;
						hit = gRadControl->radObject->IntersectsWithRay(
								srcPtV, dstPtV);
						if (!hit)
							*visPtr	+= vf;
						else
							result = 0.0;

#ifdef RAD_VIS
						if (gRadControl->showRays)
						{
							if (hit)
								gRadControl->radObject->display->C(cRed);
							else
								gRadControl->radObject->display->C(cGreen);
							gRadControl->radObject->display->P(srcPtV).P(dstPtV);
						}
#endif
							
					}
				}
			}
			samples[i][j] = result;
		}
	}

#ifdef RAD_VIS
	if ((gRadControl->showRays && visPtr) || gRadControl->showLinks)
	{
		gRadControl->radObject->display->End();
		RM_DISPLAY_END;
		RM_OUT1("Visibility: " << *visPtr);
		if (RM_PAUSE) 
			;
	}	
#endif
}


Void HRMeshElem::EstSubFormFactor(
							Int				nFrom,
							HRMeshElem		*to,
							const GCLMat	&toU,
							GCLMat			&samples, 
							GCLReal			*visPtr
						)
//	x, y E [0, 1], origin at patch's bottom left corner (vertex 1).
// XXX this is way too big and way too crufty. split up.
{
	GCLReal	rad4, npt, res2;
	GCLReal	result, vf = 1.0 / (sqr(nFrom) * toU.Rows()), xs, ys;
	Vector	temp, dxy;
	Point	srcPt[MAX_ELT_SAMPLES], srcPtV, *srcPtP;
	Vector	dstPt[MAX_ELT_SAMPLES], dstPtV;
	Vector	dx2 = to->Vertex(2) - to->Vertex(1);	
	Vector	dy2 = to->Vertex(0) - to->Vertex(1);	
	Vector	dx1 = 	  Vertex(2) -     Vertex(1);	
	Vector 	dy1 = 	  Vertex(0) -     Vertex(1);	
	GCLReal rn = nFrom;
	Int		i1, j1, i2, j;
	
#ifdef RAD_VIS
	if (gRadControl->showRays && visPtr)
		RM_DISPLAY_START;
#endif

	samples.SetSize(toU.Rows(), sqr(nFrom));
	
	// generate sample points at centers of panes in source
	srcPtP = srcPt;
	for (i1 = 0; i1 < nFrom; i1++)
		for (j1 = 0; j1 < nFrom; j1++)
		{
			if (IsQuad())
			{
				xs = (j1 + 0.5) / rn;
				ys = (i1 + 0.5) / rn;
			}
			else if (i1 + j1 >= nFrom)
			{
				xs = 1 - (i1 + 1.0/3.0) / rn;
				ys = 1 - (j1 + 1.0/3.0) / rn;
			}
			else
			{
				xs = (j1 + 1.0/3.0) / rn;
				ys = (i1 + 1.0/3.0) / rn;
			}
			*srcPtP++ = (Vertex(1) + xs * dx1 + ys * dy1);
		}

	// these will come in handy if we have to explicitly find 
	// the pane's corners
	dx1 /= rn;
	dy1 /= rn;
	dxy = -(dx1 + dy1); // vector to get from centre of pane to upper-left
	if (IsQuad())
		dxy /= 2.0;
	else
		dxy /= 3.0;

	// find destination sample points
	for (i2 = 0; i2 < toU.Rows(); i2++)
	{
		dstPt[i2] = to->Vertex(1);
		dstPt[i2] += toU[i2][0] * dx2;
		dstPt[i2] += toU[i2][1] * dy2;
	}

	for (i2 = 0; i2 < toU.Rows(); i2++)
	{
		for (srcPtP = srcPt, j = 0, i1 = 0; i1 < nFrom; i1++)
			for (j1 = 0; j1 < nFrom; j1++, srcPtP++, j++)
			{
				temp = *srcPtP;
				temp -= dstPt[i2];
		
#ifdef RAD_VIS
				if (gRadControl->showLinks)
					gRadControl->radObject->display->P(dstPt[i2]).P(*srcPtP);
#endif
			
				result = dot(to->Normal(), temp);
			
				if (result <= 0)
					result = 0.0;
				else 
				{
					npt = -dot(Normal(), temp);
				
					if (npt <= 0)
						result = 0.0;
					else
					{
						GCLReal sArea = area / sqr(rn);
					
						rad4 = sqr(sqrlen(temp));
						result *= sArea / vl_pi;
						res2 = result * Max(npt, 0.25 * sqrt(sArea));
		
						if (gRadControl->quadLevel > 0 &&
							rad4 * gRadControl->dFError < res2)
						{
							Vector v0, v1, v2, v3;
		
							// result is blowing up!
							// let's bail to the area-to-point formula
						
							// work out exact vertices of the source pane
							v1 = temp; // this is vector to its centre
							if (IsQuad())
							{
								v1 += dxy;
								v0 = v1 + dy1;
								v2 = v1 + dx1;
								v3 = v2 + dy1;
							}
							else if (i1 + j1 >= nFrom)
							{
								v1 -= dxy;
								v0 = v1 - dy1;
								v2 = v1 - dx1;
							}
							else
							{
								v1 += dxy;
								v0 = v1 + dy1;
								v2 = v1 + dx1;
							}				
										
							// find analytic A->dA factor
							result = EdgeArea(v0, v1, to->Normal());
							result += EdgeArea(v1, v2, to->Normal());
							if (IsQuad())
							{
								result += EdgeArea(v2, v3, to->Normal());
								result += EdgeArea(v3, v0, to->Normal());
#ifdef RAD_VIS
								if (gRadControl->showLinks)
									gRadControl->radObject->display->
										P(dstPt[i2] + v0).P(dstPt[i2] + v1).
										P(dstPt[i2] + v1).P(dstPt[i2] + v2).
										P(dstPt[i2] + v2).P(dstPt[i2] + v3).
										P(dstPt[i2] + v3).P(dstPt[i2] + v0);
#endif
							}
							else
							{
								result += EdgeArea(v2, v0, to->Normal());
#ifdef RAD_VIS
								if (gRadControl->showLinks)
									gRadControl->radObject->display->
										P(dstPt[i2] + v0).P(dstPt[i2] + v1).
										P(dstPt[i2] + v1).P(dstPt[i2] + v2).
										P(dstPt[i2] + v2).P(dstPt[i2] + v0);
#endif
							}
							if (result < 0.0)
								result = 0.0;
						}
						else
							result *= npt / rad4;

						if (visPtr)
						{
							Bool hit;
							
							dstPtV = dstPt[i2] + kRaySurfEps * to->Normal();
							srcPtV = *srcPtP + kRaySurfEps * Normal();
						
							hit = gRadControl->radObject->IntersectsWithRay(
								srcPtV, dstPtV);
							if (!hit)
								*visPtr	+= vf;
							else
								result = 0.0;
#ifdef RAD_VIS
							if (gRadControl->showRays)
							{
								if (hit)
									gRadControl->radObject->display->
										Begin(renLines).C(cRed).P(srcPtV).P(dstPtV).End();
								else
									gRadControl->radObject->display->
										Begin(renLines).C(cGreen).P(srcPtV).P(dstPtV).End();
							}
#endif
						}
					}
				}
				samples[i2][j] = result;
			}
	}
#ifdef RAD_VIS
	if (gRadControl->showRays && visPtr)
	{
		RM_DISPLAY_END;
		if (visPtr)
			RM_OUT1("Visibility: " << *visPtr);
		if (RM_PAUSE)
			;
	}	
#endif
}

Void HRMeshElem::SubToSubFormFactor(
							Int			nFrom, 
							HRMeshElem	*to,
							Int			nTo, 
							GCLMat		&samples, 
							GCLReal		*visPtr
						)
{
	Coord	uv[MAX_ELT_SAMPLES], *uvP;
	Int		i2, j2;
	
	Assert(sqr(nTo) < MAX_ELT_SAMPLES, "increase MAX_ELT_SAMPLES");
	// find uv samples on the destination
	if (to->IsQuad())
		for (uvP = uv, i2 = 0; i2 < nTo; i2++)
			for (j2 = 0; j2 < nTo; j2++, uvP++)
			{
				*uvP = Coord(j2 + 0.5, i2 + 0.5);
				*uvP *= 1.0 / nTo;	
			}
	else
		for (uvP = uv, i2 = 0; i2 < nTo; i2++)
			for (j2 = 0; j2 < nTo; j2++, uvP++)
			{
				if (i2 + j2 >= nTo)
				{
					*uvP = Coord(1.0 - i2 + 1.0/3.0, 1.0 - j2 + 1.0/3.0);
					*uvP *= 1.0 / nTo;	
				}
				else
				{
					*uvP = Coord(j2 + 1.0/3.0, i2 + 1.0/3.0);
					*uvP *= 1.0 / nTo;	
				}
			}
			
	// call EstSubFormFactor with the results
	EstSubFormFactor(nFrom, to, GCLMat(sqr(nTo), 2, (GCLReal*) uv), samples, visPtr);
}


// --- Drawing ----------------------------------------------------------------


Void HRMeshElem::Draw(Renderer &r)
{
	// pass on to the HRElem Draw
	DrawElem(r);
}

Void HRMeshElem::DrawSampledLeaf(Renderer &r, Int n)
{
	Int			i, j;
	Vector		dx = (Vertex(2) - Vertex(1)) / GCLReal(n);
	Vector		dy = (Vertex(0) - Vertex(1)) / GCLReal(n);
	Point		pt;
	GCLReal		ds = 1.0 / GCLReal(n);
	GCLReal		su, sv;
	Bool		project = gRadControl->funcView;
			
	// Render patch by drawing a mesh of (linearly interpolated)
	// sample points. 'n' controls the number of elements in the mesh.
	// This is a generic drawing function that only requires SampleLeaf()
	// to be defined.

	for (i = 0; i < n; i++)
		for (j = 0; j < n; j++)
		{
			// draw pane (i, j)
			if (project)
			{
				r.Begin(renPoly);

				if (IsQuad())
				{
					pt = Vertex(1) + dx * j + dy * i;
					su = ds * j; sv = ds * i;
// XXX
//					Project(r, SampleLeaf(Coord(su, sv + ds)), pt + dy);
//					Project(r, SampleLeaf(Coord(su, sv)), pt);
//					Project(r, SampleLeaf(Coord(su + ds, sv)), pt + dx);
//					Project(r, SampleLeaf(Coord(su + ds, sv + ds)), 
//							pt + dx + dy);
				}
				else if (j + i >= n)
				{
					pt = Vertex(1) + dx * (n - i) + dy * (n - j);
					su = 1 - ds * i; sv = 1 - ds * j;
//					Project(r, SampleLeaf(Coord(su, sv - ds)), pt - dy);
//					Project(r, SampleLeaf(Coord(su, sv)), pt);
//					Project(r, SampleLeaf(Coord(su - ds, sv)), pt - dx);
				}
				else
				{
					pt = Vertex(1) + dx * j + dy * i;
					su = ds * j; sv = ds * i;
//					Project(r, SampleLeaf(Coord(su, sv + ds)), pt + dy);
//					Project(r, SampleLeaf(Coord(su, sv)), pt);
//					Project(r, SampleLeaf(Coord(su + ds, sv)), pt + dx);
				}

				r.End();
			}
			else
			{
				r.Begin(renPoly);

				if (IsQuad())
				{
					pt = Vertex(1) + dx * j + dy * i;
					su = ds * j; sv = ds * i;
					r.C(SampleLeaf(Coord(su, sv + ds))).P(pt + dy);
					r.C(SampleLeaf(Coord(su, sv))).P(pt);
					r.C(SampleLeaf(Coord(su + ds, sv))).P(pt + dx);
					r.C(SampleLeaf(Coord(su + ds, sv + ds))).P(pt + dx + dy);
				}
				else if (j + i >= n)
				{
					pt = Vertex(1) + dx * (n - i) + dy * (n - j);
					su = 1 - ds * i; sv = 1 - ds * j;
					r.C(SampleLeaf(Coord(su, sv - ds))).P(pt - dy);
					r.C(SampleLeaf(Coord(su, sv))).P(pt);
					r.C(SampleLeaf(Coord(su - ds, sv))).P(pt - dx);
				}
				else
				{
					pt = Vertex(1) + dx * j + dy * i;
					su = ds * j; sv = ds * i;
					r.C(SampleLeaf(Coord(su, sv + ds))).P(pt + dy);
					r.C(SampleLeaf(Coord(su, sv))).P(pt);
					r.C(SampleLeaf(Coord(su + ds, sv))).P(pt + dx);
				}

				r.End();
			}
		}	
}

Void HRMeshElem::DrawMatrixRec(Void *dmi)
{
#ifdef RAD_VIS
	HRLinkIter	i;
	Int			j;
	GCLReal		x, y, wx, wy, n, m, temp, start, wstart;
	DMInfo		*dmInfo = (DMInfo*) dmi;
	Renderer	*r = dmInfo->r;
	Int			baseNum = dmInfo->baseNum;

	// Step through links, and draw each one in the appropriate position.
	// to do this for a general hierarchy, we would need a complete treeCode
	// for all HRElems in the hierarchy...

	// find the "square" our root patch is in
	wstart = 2.0 / baseNum;
	start = 1.0 - (2.0 * props->id) / baseNum;

	m = 1.0 / (1 << (2 * level));
	wy = wstart * m;
	y = start - treeCode * wy;	
	
	// Draw row of the matrix
	
	for (i.Begin(links); !i.AtEnd(); i.Inc())
	{
		HRMeshElem	*elt = (HRMeshElem*) i.Data().from;
		
		start = (2.0 * elt->props->id) / baseNum - 1.0;
		n = 1.0 / (1 << (2 * elt->level));
		wx = wstart * n;
		x = start + elt->treeCode * wx;
		
		if (highlight == 2 || elt->highlight == 3) 
		{
			if (highlight == 2 && elt->highlight == 3)
				r->C(cPurple);
			else if (highlight == 2) 
				r->C(cYellow);
			else
				r->C(cGreen); 
			PaintRect(*r, Coord(x, y - wy), Coord(x + wx, y));
		}
		else
			// we are drawing form-factor, which is proportional to 'from'
			// area, which is prop. to 1 / (2 ^ 2 * level)
			i.Data().DrawLink(*r, x, y - wy, x + wx, y, baseNum * 
				(1 << (2 * elt->level)));
	}

	if (!IsLeaf())
		ApplyToChildren(&HRElem::DrawMatrixRec, dmi);
#endif
}

				
Void HRMeshElem::EltSetVisPoints(HRElem *to, Point p[]) 
{
	SetVisPoints(p);
};

Void HRMeshElem::EltUpdateBounds(Point &min, Point &max)
{
	UpdateBounds(Vertex(0), min, max);
	UpdateBounds(Vertex(1), min, max);
	UpdateBounds(Vertex(2), min, max);
	if (IsQuad())
		UpdateBounds(Vertex(3), min, max);
}

Void HRMeshElem::DrawNodeElem(Renderer &r)
{
	Int i;

	if (flags.IsSet(hrMark))
	{
		flags.UnSet(hrMark);
		return;
	}

	if (HasChildren() 
#ifdef RAD_VIS
			&& !highlight
#endif
			&& (!gRadControl->choke || gRadControl->choke > level))
		for (i = 0; i < 4; i++)
			((HRMeshElem*) child[i])->DrawElem(r);
	else
		DrawLeafElem(r);
}

Void HRMeshElem::DrawLeafElem(Renderer &r)
{
	DrawLeaf(r);
}

Void HRMeshElem::HRCorrectLeaves()
{
	if (IsLeaf())
		CorrectLeaves();
	else
	{
		Child(0)->HRCorrectLeaves();
		Child(1)->HRCorrectLeaves();
		Child(2)->HRCorrectLeaves();
		Child(3)->HRCorrectLeaves();
	}
}

static GCLReal tRegSquareSamples[16][2] =
{
	0.125, 0.125,
	0.375, 0.125,
	0.625, 0.125,
	0.875, 0.125,
	0.125, 0.375,
	0.375, 0.375,
	0.625, 0.375,
	0.875, 0.375,
	0.125, 0.625,
	0.375, 0.625,
	0.625, 0.625,
	0.875, 0.625,
	0.125, 0.875,
	0.375, 0.875,
	0.625, 0.875,
	0.875, 0.875,
};

Void HRMeshElem::EltGetSamples(Int numSamples, Point pts[])
{
	Int 		i;
	Vector		dx, dy;
	Coord 		*samps;
	Point		orig, sampPt;
	
	dy =  Vertex(0) - Vertex(1);
	dx = Vertex(2) - Vertex(1);
	orig = Vertex(1);
	
	if (IsQuad())
	 	samps = (Coord *) tSquareSamples; 
	else
		samps = (Coord *) tTriangleSamples;

	for (i = 0; i < numSamples; i++)
	{
		sampPt = orig;
		sampPt += samps[i][0] * dx;
		sampPt += samps[i][1] * dy;
		pts[i] = sampPt;
	}
}

Void HRMeshElem::SetHighlight(Int h)
{
#ifdef RAD_VIS
	HRElem::SetHighlight(h);
	RadElem::SetHighlight(h);
#endif
}
