/*
	File:			HierMesh.h

	Function:		Provides a recursively-subdividable mesh of quadrilaterals
					or triangles.
					
	Author(s):		Andrew Willmott

	Copyright:		(c) 1997-2000, Andrew Willmott
 */


#ifndef __HierMesh__
#define __HierMesh__

#include "RadMesh.h"

// element capable of hierarchical subdivision...

class HierElem;
typedef Array<HierElem> HierElemList;
typedef HierElem *HierElemPtr;

class HierElem : public NbRadElem
{
public:
						HierElem();
	virtual 			~HierElem();
	
	virtual Void		SetParent(HierElem &parent);
	virtual HierElem 	*New() 
						{ return(new HierElem); };
	
	Bool				HasChildren()
						{ return(child[0] != 0);};
    Void				FreeChildren();

	Void 				Draw(Renderer &r);
	virtual Void 		DrawLeaf(Renderer &r);
	Void				DrawQuad(Renderer &r, RenderStyle start, Int code,
								 RadElem *mid);
	Void				DrawTri(Renderer &r, RenderStyle start, Int code, 
								RadElem *mid);
	Void				ProjQuad(Renderer &r, RenderStyle start, Int code,
								 RadElem *mid);
	Void				ProjTri(Renderer &r, RenderStyle start, Int code,
								RadElem *mid);
	
	Void				Reanimate(RadElem *parent);
	Void				ConnectNeighbours();
	Void 				Subdivide();

	Void				SetColour(const Colour &c);

	Void 				ColourVertices(Int weights[]);
	RadElem				*FindContainer(Coord &coord);
						// find the leaf element containing this coord.

	Void				CreatePatches(PatchList &patches);
	Void				CreateElements(PatchList &elements,
									   IndexList &eltParents, Int parent);
	GCLReal				RadError();

	// utilities
	Int					CalcMidCode(RadElem *mid);
	Void				CollectLeaves(PatchList &leaves);
	Void				SplitElement(Int &numElems, HierElemPtr	subElems[5]);
	Void				CorrectLeaves();
		
	// Printing and Parsing
	Void				Print(ostream &s);	
	Void				PrintRec(ostream &s);	
	virtual Void		PrintSelf(ostream &s);	
	Void				Parse(istream &s);	
	virtual Void		ParseSelf(istream &s);	
	Void				PrintID(ostream &s);	// Print code ID.
	StrConst			Name();

	GCLReal				MemoryUse();

	// Accessor functions
	HierElem			*&NbFace(Int i)		
						{ return((HierElem*&) nbFace[i]); };

	// Fields	
	HierElem			*child[4];			// Children, if subdivided...
	HierElem			*parent;	
	Int					level;		// level of the quadtree.
	UInt32				treeCode;	// used to encode position in quadtree.
};

/*
	Field explanations.
	
	treeCode:   Defines the path from the root of the tree to this node.
				At level i of the tree, (3 & (treeCode >> (2 * (level
				- i)))) is the child to follow.  E.g., treeCode =
				011110 for a level 3 patch: the patch is child 2 of
				child 3 of child 1 of the root.  With 32 bit integers,
				this allows for 16 different levels: the lowest will
				have 1/65535 the side length of the highest, so that
				should be plenty. 
*/

// A grid of HierElems

class HierGrid : public GridBase
{
public:
	Void			CreatePatches(PatchList &patches);
	Void 			Draw(Renderer &r);
	Void			ColourVertices(Int weights[]);
	RadElem 		*FindContainer(Coord &coord);	
	Void			Reanimate(RadElem *parent);
	
	Void			Print(ostream &s);	
	Void			Parse(istream &s);	
	StrConst		Name();

	RadElemPtr		AddChild(RadElem &quad, Int i, Int j, Int in, Int jn);
					// override.
	Void			ConnectChildren();

	GCLReal			MemoryUse();

	// Fields
	HierElemList	children;
};

#endif
