/*
	File:			MatRad.cc

	Function:		See header file

	Author(s):		Andrew Willmott

	Copyright:		(c) 1997-2000, Andrew Willmott

	Notes:			

*/

#include "MatRad.h"
#include "cl/Timer.h"
#include "gcl/Draw.h"
#include <fstream.h>
#include <stdio.h>

Matd MatRad::EmissionVectors(PatchList &patches)
{
	Matd	result(3, patches.NumItems());
	Int		i, j;

	for (i = 0; i < patches.NumItems(); i++)
		for (j = 0; j < 3; j++)
			result[j][i] = patches[i]->Emittance()[j];
		
	return(result);
}

SparseVecd MatRad::FormFactorToVector(Int j, PatchList &patches)
// Find the vector of patch factors to patch j
{
	const GCLReal	epsilon = 1e-6;
	SparseVecd		result(patches.NumItems());
	Int				i;
	GCLReal			vis;
	
#ifdef RAD_VIS
	if (gRadControl->showRays && gRadControl->visibility == vis_1)
		display->Clear().Draw(GetScene());	
#endif
	
	if (len(patches[j]->Reflectance()) <= 1e-6)
		result.MakeUnit(j);
	else
	{
		result.Begin();
		for (i = 0; i < patches.NumItems(); i++)
			if (i != j)
			{
				vis = patches[i]->Visibility(patches[j]);
			
				if (vis > 0.0)
					result.AddElt(i, patches[i]->EstFormFactor(patches[j])
								  * vis);
			}
		result.End();
	}
	
#ifdef RAD_VIS
	if (gRadControl->showRays && gRadControl->visibility == vis_1)
	{
		display->Show();
		Pause();
	}
#endif
	
	return(result);
}

Bool MatRad::Render()
//	Returns true if no interruptions.
{
	Int		i, j;
	Vec3d	error;
	Int		numPolys = patches.NumItems();
	Colour	c;

	RadMethod::Render();
	SparseVecd::SetFuzz(1e-10);
	
	SparseVecd FFRow(numPolys);	// Row of form factors
			
	for (i = 0; i < 3; i++)
		A[i].SetSize(numPolys, numPolys);
	
	E.SetSize(3, numPolys);
	B.SetSize(3, numPolys);

	numPatches = 0;

	// Flat colour scene according to current reflectances
	ColourMeshInitial();

	if (Stage(1)) return(0);
	
	E = EmissionVectors(patches);

	if (Stage(2)) return(0);

	// Initialise A's to unit matrices
	for (j = 0; j < 3; j++)
		A[j] = vl_I;

	if (Stage(3)) return(0);
																
	// Calculate solution matrices
	for (i = 0; i < patches.NumItems(); i++)	
	{
		numPatches++;
		// find form factors to this patch
		FFRow = FormFactorToVector(i, patches);	
				
		for (j = 0; j < 3; j++)
		{
			A[j][i] -= FFRow * patches[i]->Reflectance()[j];
			
			c[j] = patches[i]->Emittance()[j] + 
				patches[i]->Reflectance()[j] * dot(FFRow, E[j]);
		}
		patches[i]->SetColour(c);
		
		if (Stage(4)) return(0);
	}
		
	if (Stage(5)) return(0);
	
	error = vl_1;
		
	while (len(error) > 0.001)
	{		
		if (!gRadControl->useConjGrad)
		{
			for (i = 0; i < 3; i++)			// Solve one more iteration
				error[i] = SolveOverRelax(A[i], B[i], E[i], 1e-6,
										  gRadControl->alpha);
		}
		else
			for (i = 0; i < 3; i++)			// Solve one more iteration
				error[i] = SolveConjGrad(A[i], B[i], E[i], 1e-6);
			
		if (Stage(6)) return(0);
	}

	if (Stage(7)) return(0);
	
	return(1);
}

#ifdef RAD_VIS

Void MatRad::DumpStats() {}

Int MatRad::Stage(Int stage)
{	
	Bool	animate = gRadControl->animate || gRadControl->pause;
	Int		i, j;
	Colour	c;

	switch (stage)
	{
	case 1:		// pre setup
		Field(out3) << patches.NumItems() << " patches."  << show;
		Field(out1) << "Forming emission vector E..." << show;
		break;
		
	case 2:		// post setup		
		Field(out1) << "Flat-colouring scene..." << show;
			
		for (i = 0; i < patches.NumItems(); i++)	// distribute colour
		{
			for (j = 0; j < 3; j++)
				c[j] = patches[i]->Reflectance()[j] 
					* Max((double) 0.2, E[j][i]);
				
			patches[i]->SetColour(c);
		}
		
		ColourVertices();
		display->Redraw();		// draw
	
		Field(out1) << "1. Initialise F matrices..." << show;
		break;
		
	case 3:		
		Field(out1) << "2. Calculate F matrices..." << show;
		StartUpdate();
		break;
		
	case 4:		// After patch setup
		doUpdate = animate || Update();

		if (doUpdate)
		{
			if (gRadControl->animate && numPatches < patches.NumItems())
				patches[numPatches]->SetHighlight(1);
			ColourVertices();
			display->Redraw();
			if (gRadControl->animate && numPatches < patches.NumItems())
				patches[numPatches]->SetHighlight(0);

			RenderMatrix();
			Field(out3) << numPatches << " / " 
						<< patches.NumItems() << " patches."  << show;

			UpdateCont();
		}
		break;
		
	case 5:		// After FF calculation
		ColourVertices();
		display->Redraw();
		RenderMatrix();
		Field(out1) << "3. Solve equations..." << show;
		StartUpdate();
		solverIterations = 0;
		break;
		
	case 6:		// Middle of solve loop.
		solverIterations++;
		if (animate || Update())
		{
			// Colour scene according to current B vector
			for (i = 0; i < patches.NumItems(); i++)
			{
				for (j = 0; j < 3; j++)
					c[j] = B[j][i];
					
				patches[i]->SetColour(c);
			}

			ColourVertices();
			display->Redraw();	// Draw!
			Field(out3) << solverIterations << " iterations"  << show;
			UpdateCont();
		}
		break;
		
	case 7:
		ColourVertices();
		display->Redraw();

		Field(out3) << "Solving the matrix took " 
			<< solverIterations << " iterations" << show;
		Field(out1) << "Finished!" << show;
		break;
	}

	if (Idle()) return(1);
	
	return(Pause());
}

#else

Void MatRad::DumpMatrix()
{
	Int     i, j, k;     
	Char    *rgb = "rgb";     
	String  temp;
	FILE    *file;       
	
	for (k = 0; k < 3; k++)
	{
		temp.Printf("%s.%c.mlab", gRadControl->outFile, rgb[k]);
		file = fopen(temp, "w");
		fprintf(file, "[\n");
		for (i = 0; i < A[k].Rows(); i++)
		{
			for (j = 0; j < A[k].Cols() - 1; j++)
				fprintf(file, "%.18g, ", A[k][i][j]);
			fprintf(file, "%.18g\n", A[k][i][j]);
		}
		
		fprintf(file, "]\n");
		fclose(file);
	}

	temp.Printf("%s.stats", gRadControl->outFile);
	file = fopen(temp, "w");
	fprintf(file, "minArea    %.18g\n", stats.minArea);
	fprintf(file, "maxArea    %.18g\n", stats.maxArea);
	fprintf(file, "minRefl    [%.18g %.18g %.18g]\n",
		stats.minRefl[0], stats.minRefl[1], stats.minRefl[2]);
	fprintf(file, "maxRefl    [%.18g %.18g %.18g]\n",
		stats.maxRefl[0], stats.maxRefl[1], stats.maxRefl[2]);
	fclose(file);
}

Void MatRad::DumpStats()
{
	GCLReal density, mem;
	Int i, j;
			
	density = 0;
	for (i = 0; i < 3; i++)
		for (j = 0; j < A[i].Rows(); j++)
			density += A[i][j].NumItems();

	mem = (density * (sizeof(GCLReal) + sizeof(Int))) + 
			sizeof(GCLReal) * E.Rows() * E.Cols() +
			sizeof(GCLReal) * B.Rows() * B.Cols();
	mem /= 1024.0;
	
	density /= 3 * A[0].Rows() * A[0].Cols();
	
	cout << dumpID
		<< ' ' << totTime
		<< ' ' << gRadControl->stage
		<< ' ' << numPatches
		<< ' ' << patches.NumItems()
		<< ' ' << gRadControl->rays
		<< ' ' << density
		<< ' ' << mem
		<< ' ' << grid->MemoryUsage()
 		<< ' ' << TotalMemoryUse()
		<< endl;

	DumpScene();
}

Int MatRad::Stage(Int stage)
{
	Colour c;

	if (CheckTime()) return(1);
	// CheckTime calls DumpStats if enough time has passed, and
	// also stops the clock.

	gRadControl->stage = stage;

	switch (stage)
	{
	case 1:		// pre setup
		cout << "method mat" << endl;
		if (!gRadControl->useConjGrad)
			cout << "solver overrelaxation" << endl;
		else
			cout << "solver conjugate-gradient" << endl;

		cout << "sub " << gRadControl->patchSubdivs << endl;
		cout << "alpha " << gRadControl->alpha << endl;
		cout << "scene " << sceneName << endl;
		cout << "polys " << gRadControl->numPolys << endl;
		cout << "format "
			 << "ID "
			 << "time "
			 << "stage "
			 << "procPatches "
			 << "patches "
			 << "rays "
			 << "density " 
			 << "mem "
			 << "rtMem "
			 << "rss "
			 << endl; 
		cout << "----------------------------------------------------------" 
			 << endl;
	
		DumpStats();
		break;
				
	case 5:		// Finished calculating A[k]
		if (gRadControl->drawMatrix)
			DumpMatrix();
		break;
		
	case 6:		// Middle of solve loop.
#ifndef RAD_BEST_TIMING
		Int		i, j;

		// Colour scene according to current B vector
		for (i = 0; i < patches.NumItems(); i++)
		{
			for (j = 0; j < 3; j++)
				c[j] = B[j][i];
				
			patches[i]->SetColour(c);
		}
#endif
		break;
		
	case 7:
		DumpStats();
		break;
	}

	if (Idle()) return(0);
	timer.ContTimer();	
	return(0);
}

#endif

Void MatRad::RemoveDirect()
{
	Int		i;

#ifdef UNFINISHED
	for (i = 0; i < 3; i++)
	{
		// b = F(b - e) = (I - A)(b - e) = b - 2e + Ae
		B[i] -= E[i];
		B[i] -= E[i];
		B[i] += A[i] * E[i];
	}
#endif
}

Void MatRad::DrawMatrix(Renderer &r)
{
#ifdef RAD_VIS
   	SVIterd		iter0, iter1, iter2;

	if (!gRadControl->drawMatrix || gRadControl->stop)
		return;
		
	Int			i, j, n;
	GCLReal		x1, x2, y1, y2, w;
		
	w = 2.0 / A[0].Cols();
	
	y1 = 1 - w;
	y2 = 1;
	n = Min(A[0].Rows(), (Int) numPatches);

	for (i = 0; i < n; i++)
	{
		x1 = -1;
		x2 = w - 1;
		
		iter0.Begin(A[0][i]);
		iter1.Begin(A[1][i]);
		iter2.Begin(A[2][i]);

		for (j = 0; j < A[0].Cols(); j++)
		{
			iter0.Inc(j);
			iter1.Inc(j);
			iter2.Inc(j);

			if (patches[i]->highlight == 2 && patches[j]->highlight == 3)
				r.C(cPurple);
			else if (patches[i]->highlight == 2)
				r.C(cYellow);
			else if (patches[j]->highlight == 3)
				r.C(cGreen);
			else if (i == j)
				r.C(cBlack);
			else	
			{
				Colour mClr = cBlack;
				
				if (iter0.Exists(j)) mClr[0] = -iter0.Data();
				if (iter1.Exists(j)) mClr[1] = -iter1.Data();
				if (iter2.Exists(j)) mClr[2] = -iter2.Data();

			   	r.C(A[0].Cols() * mClr);
			}

			PaintRect(r, Coord(x1, y1), Coord(x2, y2));
				
			x1 += w;
			x2 += w;
		}
		
		y1 -= w;
		y2 -= w;
	}
#endif
}

RadElem *MatRad::NewMesh()
{
	return(new RadGrid());
}
