/*
	File:		MultiWavelets.h

	Function:	Implements wavelet bases M2 and M3

	Author:		Andrew Willmott

    Copyright:  (c) 1996-2000, Andrew Willmott
*/
 
#ifndef __MultiWavelets__
#define __MultiWavelets__
 

#include "HRMesh.h"
#include "WaveSupport.h"


// --- M2 classes: 2 x 2 multiwavelets. (Bi-linear) ---------------------------

class M2Elem;

class M2Link : public HRLink
{
public:
	HRLink			*New() { return new M2Link; };
	
	M2Elem			*To()	{ return((M2Elem*) to); };
	M2Elem			*From()	{ return((M2Elem*) from); };

	Void			Gather();
	Bool			CalcTransport();
	Bool			CalcVisibility(HRLink *parent, Bool forceVisReuse);
	GCLReal			Error();
	GCLReal			BFAError();

	Void			Print(ostream &s);
	Void			DrawLink(Renderer &r,
							 GCLReal left, GCLReal top, 
							 GCLReal right, GCLReal bottom, 
							 GCLReal weight
					);

	// M2-specific
	GCLReal			OracleError(const GCLMat &K);
	GCLReal			CCOracleError(const GCLMat &K);
	GCLReal			SimpleOracleError(const GCLMat &K);
	GCLReal			CalcKernel(GCLMat &K, GCLReal *visPtr, Bool &usingCCRule);

	// Fields
	GCLReal			transport[16];
	GCLReal			factor;
};


class M2Elem : public HRMeshElem
{
public:
	Void		SetParent(HierElem &parent);
	StrConst	Name() { return("m2"); };
	HierElem 	*New() { return(new M2Elem); };

	Colour		SampleLeaf(Coord c);
	Void		Print(ostream &s);
	Void		PrintSelf(ostream &s);
	Void		ParseSelf(istream &s);
	
	Void		DistributeColours();
		
	Void		Add();
	Void		Push();
	Void		Pull();
	Void		CalcLeafRadiosity();
	GCLReal		Error();
	Void		ClearB();
	Void		ClearR();

	Void		DrawLeaf(Renderer &r);
	Int			Coeffs() { if (IsQuad()) return(4); else return(3); };

	// Generic wavelet interface
	Void		EltSampleTransport(Int numSamples, Point p[], Vector n[],
								   GCLMat &coeffs);
	GCLReal		EltCalcTransport(HRElem *from, GCLMat&coeffs);
	Void		AddIrradiance(const Colour &E, const Vector &m);
	Colour		GetPower(const Vector &m);

	Int			NumCoeffs() { if (IsQuad()) return(4); else return(3); };
	Colour		*B_Coeffs() { return(B); };
	Colour		*R_Coeffs() { return(R); };

	// Fields
	Colour		B[4];
	Colour		R[4];
	static Colour lastB[4];
};

// --- M3 classes: 3 x 3 multiwavelets. (Bi-quadratic) ------------------------

class M3Elem;

class M3Link : public HRLink
{
public:
	HRLink			*New() { return new M3Link; };

	M3Elem			*To()	{ return((M3Elem*) to); };
	M3Elem			*From()	{ return((M3Elem*) from); };

	Void			Gather();
	Bool			CalcTransport();
	Bool			CalcVisibility(HRLink *parent, Bool forceVisReuse);
	GCLReal			Error();
	GCLReal			BFAError();

	Void			Print(ostream &s);
	Void			DrawLink(Renderer &r,
							 GCLReal left, GCLReal top, 
							 GCLReal right, GCLReal bottom, 
							 GCLReal weight
					);

	// M3-specific
	GCLReal			OracleError(const GCLMat &K);
	GCLReal			CCOracleError(const GCLMat &K);
	GCLReal			SimpleOracleError(const GCLMat &K);
	GCLReal			CalcKernel(GCLMat &K, GCLReal *visPtr, Bool &usingCCRule);

	// Fields
	GCLReal			transport[81];
	GCLReal			factor;
};


class M3Elem : public HRMeshElem
{
public:
	Void		SetParent(HierElem &parent);
	StrConst	Name() { return("m3"); };
	HierElem 	*New() { return(new M3Elem); };

	Colour		SampleLeaf(Coord c);
	Void		Print(ostream &s);
	Void		PrintSelf(ostream &s);
	Void		ParseSelf(istream &s);
	
	Void		DistributeColours();
		
	Void		Add();
	Void		Push();
	Void		Pull();
	Void		CalcLeafRadiosity();
	GCLReal		Error();
	Void		ClearB();
	Void		ClearR();

	Void		DrawLeaf(Renderer &r);
	Int			Coeffs() { if (IsQuad()) return(9); else return(6); };

	// Generic wavelet interface
	Void		EltSampleTransport(Int numSamples, Point p[], Vector n[],
								   GCLMat &coeffs);
	GCLReal		EltCalcTransport(HRElem *from, GCLMat&coeffs);
	Void		DistributeR(const Colour &srcR, const Vector &r);

	Int			NumCoeffs() { if (IsQuad()) return(9); else return(6); };
	Colour		*B_Coeffs() { return(B); };
	Colour		*R_Coeffs() { return(R); };

	// Fields
	Colour		B[9];
	Colour		R[9];
	static Colour lastB[9];
};

#endif
