/*
	File:			ProgRad.cc

	Function:		See header file

	Author:         Andrew Willmott

	Notes:			
*/

#include "ProgRad.h"
#include "gcl/Draw.h"

/*! \class ProgRad

	This implements progressive radiosity *without* substructuring. Its
	main purpose is as a more easily readable implementation of
	progressive radiosity, without the substructuring code cluttering
	things up, and for comparisons with the substructuring version to
	gauge the overhead of subtructuring.
	
	For most "real world" applications, progressive radiosity with
	substructuring (implemented by HProgRad) should be used instead.

*/
					  
Void ProgRad::MakeFormFactorFromVector(RadElem &fromPatch, 
	PatchList &patches, ColourList &result)
// Find the vector of patch factors from patch j. 
// The factors are weighted by the reflectance of the patch they are to.
{
	Int			i, k;
	GCLReal 	visibility;
	GCLReal		factor;
	RadElem		*curPatch;
		
	result.SetSize(patches.NumItems());
		
#ifdef RAD_VIS
	if (gRadControl->showRays)
		display->Clear().Draw(GetScene());
#endif
	
	for (i = 0; i < patches.NumItems(); i++)		
	{	
		GCLReal rho;
				
		curPatch = patches[i];
		
		rho = len(curPatch->Reflectance());
		
		if (rho < 0.0001)
		{
			result[i] = vl_0;
			continue;
		}

		factor = fromPatch.EstFormFactor(curPatch);
#ifdef RAD_SPEEDY
		if (rho * factor < 1e-5)
		{
			result[i] = vl_0;
			continue;
		}
#endif

		visibility = fromPatch.Visibility(curPatch);
		
		if (visibility > 0.0001)
		{
			result[i] = curPatch->Reflectance();
			result[i] *= factor * visibility;
		}
		else
			result[i] = vl_0;
	}
	
#ifdef RAD_VIS
	if (gRadControl->showRays)
		display->Show();
#endif
}

Bool ProgRad::Render()
//	Returns true if no interruptions.
{
	Int			numPatches = patches.NumItems();
	Int 		i, j;
	
	Colour		deltaRad;
	Colour		radToShoot;
	GCLReal		power, envArea;
	Colour		envPower, envRefl;
	Bool		finished = 0;	
	
	RadMethod::Render();

	iterations = 0;
	error = 1.0;
	origShoot = 0;
		
	B.SetSize(numPatches);
	S.SetSize(numPatches);

	envRad = vl_1;
			
	if (Stage(1)) return(0);
	
	for (i = 0; i < patches.NumItems(); i++)
		S[i] = patches[i]->Emittance();
	
	B = S;
				
	//	Setup for the ambient term

	envRefl.MakeZero();
	envArea = 0;
	
	for (i = 0; i < numPatches; i++)
	{
		envRefl += patches[i]->area * patches[i]->Reflectance();
		envArea += patches[i]->area;	
	}
	
	envRefl /= Colour(vl_1) / (Colour(vl_1) - envRefl / envArea);
	// R = 1 / (1 - avgReflectance)
	
	//	Now, run the algorithm...

	while (!finished)
	{	
		if (Stage(2)) return(0);

		iterations++;

		maxPower = 0;
		maxPowerIndex = 0;
		envPower.MakeZero();
		
		for (i = 0; i < numPatches; i++)						
		{
			power = dot(S[i] * patches[i]->area, kRadRGBToLum);
				
			if (maxPower < power)
			{
				maxPower = power;
				maxPowerIndex = i;
			}
		}
		
		if (Idle())	
			return(0);

		if (Stage(3)) return(0);
			
		radToShoot = S[maxPowerIndex];
		radToShoot *= gRadControl->alpha;
		S[maxPowerIndex] -= radToShoot;
				
		if (Stage(4)) return(0);
				
#ifdef RAD_VIS
		if (gRadControl->drawMatrix)
			MakeFormFactorFromVector(*patches[maxPowerIndex], patches,
					FFRows[maxPowerIndex]);
		else			
#else
		MakeFormFactorFromVector(*patches[maxPowerIndex], patches, FFRow);
#endif

		if (Stage(5)) return(0);
		
		error = 0;
		envPower.MakeZero();
				
		for (i = 0; i < numPatches; i++)
		{
#ifdef RAD_VIS
			if (gRadControl->drawMatrix)
				deltaRad = FFRows[maxPowerIndex][i] * radToShoot;
			else
#endif
			deltaRad = FFRow[i] * radToShoot;
			B[i] += deltaRad;
			S[i] += deltaRad;
			error += sqrlen(deltaRad);

			envPower += S[i] * patches[i]->area;
		}
		
		if (gRadControl->ambient)
			envRad = envRefl * envPower / envArea;
		else
			envRad.MakeZero();

		error = sqrt(error);

		if (iterations == 1)
			origShoot = maxPower;
		else if (maxPower < gRadControl->error * origShoot)
			finished = 1;
		
		if (Stage(6)) return(0);
	}	
	
	if (Stage(7)) return(0);
	
	return(1);
}

#ifdef RAD_VIS

Void ProgRad::DumpStats(){}

Int ProgRad::Stage(Int stage)
{
	Int		i, j;
	Bool	animate = gRadControl->animate || gRadControl->pause;

	switch (stage)
	{
	case 1:		// pre setup
		Field(out3) << patches.NumItems() << " patches"  << show;
		Field(out1) << "Forming emission vector E..." << show;
		FFRows.Clear();
		if (gRadControl->drawMatrix)
		{
			FFRows.SetSize(patches.NumItems());
//			FFRows.ClearTo(ColourList());
		}
		StartUpdate();
		break;
	case 2:		// post setup
		ColourVertices();
		display->Redraw();

		doUpdate = animate || Update();

		if (animate)
			Field(out1) << "Finding patch to shoot..." << show;
		else if (doUpdate)
		{
			Char *plural;

			if (iterations == 1)
				plural = "";
			else
				plural = "es";

			Field(out1) << iterations << " patch" << plural << " shot" << show;

			RenderMatrix();
		}
		break;
		
	case 3:		
		if (animate)
			Field(out1) << "Patch " << maxPowerIndex << " next to shoot"
						<< show;
		break;
		
	case 4:		// After patch setup
		if (doUpdate)
		{
 			if (animate)		   
			{
				Field(out1) << "Calculating form factors..." << show;
				// flash shooting patch red
				patches[maxPowerIndex]->SetHighlight(1);
			}

			ColourVertices();
			
			display->Redraw();
			patches[maxPowerIndex]->SetHighlight(0);
		}
		break;
		
	case 5:	
		// Update our display matrix...
		if (doUpdate)
		{
			RenderMatrix();
			if (animate)
				Field(out1) << "Shooting Patch " << maxPowerIndex << show;
		}
		break;
		
	case 6:		// Middle of solve loop.
			
		if (doUpdate)
		{
			// Colour the patches...
			for (i = 0; i < patches.NumItems(); i++)
			{
				// Show the shot vector or the radiosity vector?
	
				if (gRadControl->shotDisplay)					
					patches[i]->colour = S[i];
				else
					patches[i]->colour = B[i] + patches[i]->Reflectance() 
						* envRad;
			}
	
			ColourVertices();
			display->Redraw();
			
			Field(out2) << "error = " << maxPower / origShoot
						<< ", RMS Error = " << error << show;
			UpdateCont();
		}
		break;
		
	case 7:
		// Colour the patches...
		for (i = 0; i < patches.NumItems(); i++)
			if (gRadControl->shotDisplay)					
				patches[i]->colour = S[i];
			else
				patches[i]->colour = B[i] + patches[i]->Reflectance() * envRad;

		ColourVertices();
		display->Redraw();
		
		Field(out2) << "error = " << maxPower / origShoot << ", RMS Error = "
					<< error << show;
		Field(out1) << "Finished: " << iterations << " patches shot" << show;
		break;
	}
	return(Pause());
}

#else

Void ProgRad::DumpStats()
{
	Int		i;
	GCLReal	shotErr, mem;

	if (origShoot == 0.0)
		shotErr = 1.0;
	else
		shotErr = maxPower / origShoot;
	
	mem = sizeof(Colour) * (B.NumItems() + S.NumItems() + FFRow.NumItems());
	mem /= 1024.0;
	
	cout << dumpID 
		<< ' ' << totTime 
		<< ' ' << gRadControl->stage 
		<< ' ' << patches.NumItems() 
		<< ' ' << shotErr
		<< ' ' << error
		<< ' ' << gRadControl->rays
		<< ' ' << iterations 
		<< ' ' << mem
		<< ' ' << grid->MemoryUsage()
		<< ' ' << TotalMemoryUse()
		<< endl;

	for (i = 0; i < patches.NumItems(); i++)		// Colour the patches...
	{
		// Show the shot vector or the radiosity vector?

		if (gRadControl->shotDisplay)			
			patches[i]->colour = S[i];
		else
			patches[i]->colour = B[i] + patches[i]->Reflectance() * envRad;
	}

	DumpScene();
}

Int ProgRad::Stage(Int stage)
{
	if (CheckTime()) return(1);

	gRadControl->stage = stage;

	switch (stage)
	{
	case 1:		// pre setup
		cout << "method prog " << endl;
		cout << "sub " << gRadControl->patchSubdivs << endl;
		cout << "tErr " << gRadControl->error << endl;
		cout << "scene " << sceneName << endl;
		cout << "polys " << gRadControl->numPolys << endl;
		cout << "srcPatches " << patches.NumItems() << endl;
		cout << "format "
			 << "ID "
			 << "time "
			 << "stage "
			 << "patches "
			 << "shootErr "
			 << "resErr "
			 << "rays "
			 << "iterations "
			 << "mem "
			 << "rtMem "
			 << "rss "
			 << endl;

		cout << "----------------------------------------------"
			 << "---------------------------------" << endl;
	
		DumpStats();
		break;
		
	case 7:
		DumpStats();
		break;
	}

	if (Idle()) return(0);
	timer.ContTimer();
	return(0);
}

#endif

Void ProgRad::DrawMatrix(Renderer &r)
{
	if (!gRadControl->drawMatrix || gRadControl->stop)
		return;
		
#ifdef RAD_VIS
	Int		i, j;
	GCLReal	x1, x2, y1, y2, wx, wy;
		
	wx = 2.0 / FFRows.NumItems();
	
	x1 = -1;
	x2 = wx - 1;
	
	for (i = 0; i < FFRows.NumItems(); i++)
	{
		wy = 2.0 / FFRows[i].NumItems();
		y1 = 1 - wy;
		y2 = 1;
		
		for (j = 0; j < FFRows[i].NumItems(); j++)
		{
			if (patches[j]->highlight == 2 && patches[i]->highlight == 3)
				r.C(cPurple);
			else if (patches[j]->highlight == 2)
				r.C(cYellow);
			else if (patches[i]->highlight == 3)
				r.C(cGreen);
			else	
				r.SetColour(FFRows.NumItems() * FFRows[i][j]);
			PaintRect(r, Coord(x1, y1), Coord(x2, y2)); 

			y1 -= wy;
			y2 -= wy;
		}
		
		x1 += wx;
		x2 += wx;
	}
#endif
}


RadElem *ProgRad::NewMesh()
{
	return(new RadGrid());
}
