/*
	File:			RadDraw.cc

	Function:		Renders checkpoint radiosity scene files.

	Author(s):		Andrew Willmott

	Copyright:		(c) 1997-2000, Andrew Willmott

	Notes:			

*/

#include <unistd.h>
#include <fstream.h>
#include "cl/ArgParse.h"

#include "RadMethod.h"
#include "RadScene.h"

#include "gcl/XGraphicsSystem.h"
#include "gcl/Avars.h"
#include "gcl/ScenePane.h"
#include "gcl/Readers.h"
#include "gcl/EPSRenderer.h"
#include "cl/String.h"


// ----------------------------------------------------------------------------


static Bool gDump = 0;
static Bool gMove = 0;

class RadDrawPane : public ScenePane
{
public:
	Void	HandleKey(Char c); 
};

extern RenderStyle scRenderMode;

Void RadDrawPane::HandleKey(Char c)
{
	switch (c)
	{
	case 'g':
		gRadControl->gouraud = !gRadControl->gouraud;
		Redraw();
		break;
	case 'w':
		gRadControl->wire = !gRadControl->wire;		
		if (gRadControl->wire)
			scRenderMode = renLineLoop;
		else
			scRenderMode = renPoly;
		Redraw();
		break;
	case 'm':
		gRadControl->anchor = !gRadControl->anchor;
		Redraw();
		break;
	case 'f':
		gRadControl->funcView = !gRadControl->funcView;
		Redraw();
		break;	
	case 'd':
		gDump = 1;
		break;
	case 'e':
		gDump = 2;
		break;
	case 'c':
		gDump = 3;
		break;
	case 'D':
		gDump = 4;
		break;
	case ' ':
		gMove = 1;
		break;
	case ',':
		gMove = 2;
		break;
	case '.':
		gMove = 1;
		break;
	case 'q':
		xgs->SignalDone();
		break;
	default:
		ScenePane::HandleKey(c);
	}
}

// ----------------------------------------------------------------------------

class ProgramOptions
{
public:
	char	**files;
	int		numFiles;
	char	*outFile;
	Int		size;
	Int		window;
	Int		ppm;
	Int		tiff;
	Int		eps;
	Int		trace;
	Int		final;
	Int		noFork;
	double	grey;
};

static char** gArgv;
static int gArgc;

Void GetFiles(int argc, char **argv)
{
	// Save away a pointer to the list of file names
	// I've checked the source of libarg.c -- it
	// doesn't free anything, so this is safe.

	gArgv = argv;
	gArgc = argc;
}

Void SetOptions(int argc, char **argv, RadControl &radOpts, ProgramOptions &progOpts)
{
	Int gouraud, wire, func, anchor;
	Arg_form *arg_format;
	
	progOpts.size = 400;
	progOpts.outFile = 0;
	progOpts.grey = 0.5;
		
	arg_format = arg_to_form(
			"", 										"usage: raddraw file1 file2 ... [options]",
			
			"", ARG_SUBR(GetFiles), 						"",
			"-o %S", &progOpts.outFile, 				"output file",
			"-tiff", ARG_FLAG(&progOpts.tiff),			"save as .tiff (default)",
			"-ppm", ARG_FLAG(&progOpts.ppm),			"save as .ppm",
			"-eps", ARG_FLAG(&progOpts.eps),			"save as .eps",
			
			"-gouraud", ARG_FLAG(&gouraud),				"use gouraud shading",
			"-anchor", ARG_FLAG(&anchor),				"use anchored mesh",
			"-wire", ARG_FLAG(&wire), 					"display wireframe",
			"-func", ARG_FLAG(&func), 					"function view",
			"-bg %F", &progOpts.grey, 					"Background grey level",

			"-noFork", ARG_FLAG(&progOpts.noFork), 		"don't go into background",
			"-trace", ARG_FLAG(&progOpts.trace), 		"render using ray tracer",
			"-final", ARG_FLAG(&progOpts.final), 		"render using final gather",
			"-samples %F", &radOpts.sampleFactor, 		"FG sample factor",
			
			// Output options
			
			"-window", ARG_FLAG(&progOpts.window), 		"display output in window",
			"-size %d", &progOpts.size, 				"set window size",
		0);

	// Do arg parsing
	
	if (argc == 1)
	{
		arg_form_print(arg_format);
		SceneReader::PrintSupportedFormats(cout);
		exit(1);
	}
	if (arg_parse_argv(argc, argv, arg_format) < 0)
		exit(1);

	radOpts.gouraud = gouraud;
	radOpts.anchor = anchor;
	radOpts.wire = wire;
	radOpts.funcView = func;

	if (!progOpts.eps && !progOpts.ppm)
		progOpts.tiff = 1;
	if (progOpts.final)
		progOpts.trace = 1;

	progOpts.files = gArgv;
	progOpts.numFiles = gArgc;

	if (strcmp("radview", argv[0]) == 0)
		progOpts.window = 1;
}

#include "Parse.h"


Void DrawRadScene(ProgramOptions &progOpts, RadControl &radOpts)
{
	XGraphicsSystem			xgs;
	Int						i;
	scScenePtr				scene;
	ifstream				inFile;
	GCLReal					err;
	Renderer				*display;
	RGBAImage				image;
	FileName				filename;
	RadDrawPane				*displayPane;
	Bool					done = false;
	Bool					finished;
	EPSRenderer				*epsRenderer;
	
	slInit();
	slAddToLibrary("mesh", new scRadMesh);
	
	if (progOpts.window)
	{
		if (!progOpts.noFork && fork())
			exit(0);

		cout 
			<< "keys: " << endl
			<< "  w - wireframe" << endl
			<< "  g - gouraud" << endl
			<< "  m - anchored mesh" << endl
			<< "  f - function view" << endl
			<< "  c - dump camera" << endl
			<< "  d - dump picture" << endl
			<< "  e - dump eps file" << endl
			<< "  , - previous file" << endl
			<< "  . - next file" << endl
			<< "  q - quit" << endl
			;

		displayPane = new RadDrawPane;
		display = displayPane;
		xgs.CreateWindow(displayPane, progOpts.files[0], progOpts.size,
						progOpts.size);
	}
	else if (progOpts.eps)
	{
		epsRenderer = new EPSRenderer;
		display = epsRenderer;
	}
	else if (!progOpts.trace)
	{
		GSOffscreenPane *oPane =  new GSOffscreenPane;

		xgs.CreateOffscreen(oPane, progOpts.size, progOpts.size);
		display = oPane;
	}
	
	// Window mode
	
	if (progOpts.window)
	{
		i = 0;
		finished = false;
		
		while (!finished)
		{
			filename.SetPath(progOpts.files[i]);
			
			cout << "viewing file: " << filename.GetPath() << endl;
			
			scene = SceneReader::Load(filename);
			if (!scene)
			{
				cout << "failed to read scene file " << 
					filename.GetPath() << endl;
				exit(1);
			}

			// Normalise the scene's size
			//scene->Normalise();
			displayPane->SetScene(scene);
			displayPane->SetBgColour(cWhite * progOpts.grey);
			
			// display scene
			
			while (!finished && !gMove)
			{
				xgs.Spin();
				
				if (gDump)
				{
					// Write out image/eps file...
				
					if (gDump == 4)
					{
						ofstream	fout;
						
						cout << "*** saving scene to dump.sl" << endl;
                        fout.open("dump.sl");
                        if (fout)
                        {
							fout << displayPane->ItsScene() << endl;
							fout.close();
						}
						else
							perror("Couldn't open output file");
 					}
					if (gDump == 3)
					{
						ofstream	fout;
						
						cout << "*** saving camera to view.cam" << endl;
                        fout.open("view.cam");
                        if (fout)
                        {
							((scCamera*) displayPane->
							 ItsCamera())->Print(fout);
							fout.close();
						}
						else
							perror("Couldn't open output file");
 					}
					else if (gDump == 2)
					{
						EPSRenderer eps;
						
						filename.SetExtension("eps");
						eps.Attach(filename.GetPath());
						eps.Clear().Draw(scene).Show();
					}
					else
					{
#ifdef BROKEN
						if (progOpts.trace)
						{
							image.SetSize(progOpts.size, progOpts.size);
							if (progOpts.final)
								FinalGather(scene, image, radOpts);
							else
								RayCastScene(scene, image);
						}
						else
#endif
							display->GetImage(image);
	
						if (progOpts.outFile)
							image.SavePPM(progOpts.outFile);
						else
						{
							filename.SetExtension("ppm");
							image.SavePPM(filename.GetPath());
						}
					}
					gDump = 0;
				}
			}
			
			if (gMove)
			{
				if (gMove > 1)
				{
					i--;
					if (i < 0)
						i = progOpts.numFiles - 1;
				}
				else
				{
					i++;
					if (i >= progOpts.numFiles)
						i = 0;
				}
				
				gMove = 0;
			}
		}

		return;
	}
	
	// Batch-render mode.
	
	for (i = 0; i < progOpts.numFiles; i++)
	{
		filename.SetPath(progOpts.files[i]);
		
		cout << "file: " << filename.GetPath() << endl;
		
		scene = SceneReader::Load(filename);
		if (!scene)
		{
			cout << "failed to read scene file " << filename.GetPath() << endl;
			exit(1);
		}
				
		// change output filename if we have override...
		if (progOpts.outFile)
			filename.SetPath(progOpts.outFile);

		// draw scene
			
		if (progOpts.eps)
		{
			filename.SetExtension("eps");
			epsRenderer->Attach(filename.GetPath());
		}
					
#ifdef BROKEN
		if (progOpts.trace)
		{
			image.SetSize(progOpts.size, progOpts.size);
			if (progOpts.final)
				FinalGather(scene, image, radOpts);
			else
				RayCastScene(scene, image);
		}
		else
#endif
		{
			display->SetBgColour(cWhite * progOpts.grey);
			display->Clear().Draw(scene).Show();
			if (!progOpts.eps)
				display->GetImage(image);
		}
		
		// Write out image file...

		if (progOpts.eps)
			;	// already done.
		if (progOpts.ppm)
		{
			filename.extension = "ppm";
			image.SavePPM(filename.GetPath().Ref());
		}
		if (progOpts.tiff)
		{
			filename.extension = "ppm";
			image.SavePPM(filename.GetPath().Ref());
		}
	}
}

main(int argc, char **argv)
{
	RadControl			radOpts;
	ProgramOptions		progOpts;
	
	SetOptions(argc, argv, radOpts, progOpts);

	gRadControl = &radOpts;

	if (progOpts.numFiles < 1)
	{
		exit(1);
	}
	
	if (progOpts.outFile && progOpts.numFiles > 1)
	{
		cerr << "you can only use -o with one file argument." << endl;
		exit(1);
	}
	
	DrawRadScene(progOpts, radOpts);

	return(0);
}
