/*  
	File:			RadVis.cc

	Function:		GUI front end for the various radiosity methods. Allows 
					visualisation of the methods.
					
	Author(s):		Andrew Willmott

	Copyright:		(c) 1997-2000, Andrew Willmott
 */



#include "TestScene.h"

#include "MatRad.h"
#include "ProgRad.h"
#include "HProgRad.h"
#include "HierRad.h"

#include "RadPane.h"
#include "RadScene.h"
#include "Examiner.h"

#include "gcl/XGraphicsSystem.h"
#include "gcl/Forms.h"
#include "gcl/SceneLang.h"
#include "gcl/Avars.h"
#include "gcl/Readers.h"
#include "gcl/EPSRenderer.h"
#include "gcl/MesaRenderer.h"

#include <stdlib.h>
#include <unistd.h>
#include "cl/ArgParse.h"

// classes derived from fdesign.
#include "theForms.h"	

class RadForm;

// --- exponential slider -----------------------------------------------------

// provides exponential-based slider more suited to error metrics
// basically steps down through 0.9 .. 0.1, 0.09 .. 0.01, etc.

static double eslider_map(double val)
{
	double d = floor(val - 0.05);

	return(pow(10.0, d + 1.0) * (val - d));
}

static const char *esliderFilt(FL_OBJECT *, double val, int prec)
{
	static char result[32];

	val = eslider_map(val);
	sprintf(result, "%.*g", prec, val);

	return(result);
}

double fl_get_eslider_value(FL_OBJECT *obj)
{
	return(eslider_map(fl_get_slider_value(obj)));
}

void fl_set_eslider_value(FL_OBJECT *obj, double val)
{
}

// --- Options Panels --------------------------------------------------------


class TransportPanel : public TransportForm
{
public:
	TransportPanel(RadForm *radform) : TransportForm(), parent(radform) {};
	
	Void 	Event(FL_OBJECT *object);
		
	RadForm	*parent;
};

class RenderPanel : public RenderForm
{
public:
	RenderPanel(RadForm *radform) : RenderForm(), parent(radform) {};
	
	Void 	Event(FL_OBJECT *object);
		
	RadForm	*parent;
};


// --- AvarPanel --------------------------------------------------------------


class AvarPanel : public AvarForm
{
public:
	Void 	Init();
	Void	Event(FL_OBJECT *object);
	Void	SetScene(scScenePtr scene);
	Void	Dump(ostream &s);
	
	scAvarList	*avarList;
	ScenePane	*scenePane;
};

Void AvarPanel::Init()
{
	fl_set_slider_bounds(avarSlider, 0, 1);	
	fl_set_slider_value(avarSlider, 1);	
}

Void AvarPanel::SetScene(scScenePtr scene)
{
	Int i;
	CreateAvarList avarGrabber;
	
	scene->ApplyAction(avarGrabber);
	avarList = avarGrabber.avarList;
	
	fl_clear_browser(avarBrowser);
	
	for (i = 0; i < avarList->NumItems(); i++)
		fl_add_browser_line(avarBrowser, (*avarList)[i].name);
}

Void AvarPanel::Dump(ostream &s)
{
	Int i;
	
	for (i = 0; i < avarList->NumItems(); i++)
		cout << (*avarList)[i].name << " = " << (*avarList)[i].value << endl;
}

Void AvarPanel::Event(FL_OBJECT *object)
{
	if (object == avarBrowser)
	{
		if (fl_get_browser(object) > 0)
		{
			Avar &avar = (*avarList)[fl_get_browser(avarBrowser) - 1];
			
			fl_set_slider_bounds(avarSlider, avar.upperBound,
								 avar.lowerBound);	
			fl_set_slider_value(avarSlider, avar.value);	
		}
	}
	else if (object == avarSlider)
	{
		if (fl_get_browser(avarBrowser) > 0)
		{
			(*avarList)[fl_get_browser(avarBrowser) - 1].value = 
				fl_get_slider_value(avarSlider);
			if (scenePane)
				scenePane->Redraw();	
		}
	}
	else if (object == dismiss)
		Hide();
	else if (object == dump)
		Dump(cout);
}


// --- The radvis main pane! --------------------------------------------------


class RadForm : public RadVisMainForm
{
public:

	RadForm();
		
	Int                     Main(Int argc, Char **argv);
	Void                    InitOptions(Int argc, Char **argv, 
   									RadControl &options, Char **scenePathP);

	Void					Event(FL_OBJECT *object);
	Void					Init(const Char *scenePath);
	Void					SetupControls();
	Void					SetupMenus();
	Void					SetFormOptions();
							// set form controls from options field
	Void 					GetFormOptions();
							// set options field from form controls 

	Void 					SamplePatch(Matd &samples);
	Void 					SampleFF(Matd &samples);
	Void 					UpdateMethodOptions(Int a, Int b);
	Void					RenderMode(Bool on);

	Void					UpdateRadScene();
	Void					UpdateScene();
	scScenePtr				LoadScene();
	scScenePtr				LoadScene(const Char *path);

	// Commands

	Void					Go();
	Void					Reset();
	Void					ChooseMethod();
	Void				    SceneSelect();
	Void				    DumpImage();
	Void				    DumpEPS();
	Void				    DumpObjFile();
	Void				    DumpSLFile();
	Void				    DumpCamera();
	Void					StartHelp();

	// Fields
	
	Int						oldMethodChoice;
		
	scScenePtr				scene;
	scScenePtr 				matrixScene;
	RadMethod				*radMethod;
	RadControl				options;
	
	RadScenePane			mainPane;
	ScenePane2D				matrixPane;

	AvarPanel				avarPanel;
	TransportPanel			transPanel;
	RenderPanel				renderPanel;
	
	FormsGraphicsSystem		*fgs;
	Bool					sceneSet;
	Colour					bgColour;
};


// --- RadMethod form methods -------------------------------------------------------


RadForm::RadForm() : 
	RadVisMainForm(),
	scene(0), 
	transPanel(this), 
	renderPanel(this), 
	fgs(0)
{ 
#ifdef DEBUG
	title = "Radiosity Visualisation (Debug)";
#else
	title = "Radiosity Visualisation";
#endif
	gRadControl = &options;
}

Void RadForm::Init(const Char *scenePath)
{	
	fgs->CreateFormsPane((XEventPane*) &mainPane, box1);
	fgs->CreateFormsPane(&matrixPane, box2);
	
	SetupControls();
	SetupMenus();
	SetFormOptions();
			
	//	Enable/Disable the right controls.

	UpdateMethodOptions(kMatrix, kNoMethod);
	UpdateMethodOptions(kProgressive, kNoMethod);
	UpdateMethodOptions(kProgSubstruct, kNoMethod);
	UpdateMethodOptions(kHierarchical, kNoMethod);
	UpdateMethodOptions(kAnalytical, kNoMethod);
	UpdateMethodOptions(kNoMethod, options.method);
	
	slInit();	
	if (scenePath)
		scene = LoadScene(scenePath);
	else
	{
		options.drawMatrix = true;
		options.cluster = false;
		scene = BoxScene();
	}
	
	// Set up the radiosity method and the view pane.
	
	mainPane.SetBgColour(bgColour);
	mainPane.SetParent(this);
	sceneSet = false;
	// get into 'reset' state...
	Reset();

	oldMethodChoice = 0;
	radMethod = RadMethod::NewRadMethod();
	radMethod->SetDisplay(&mainPane, fgs);

	radMethod->matDisplay = &matrixPane;
	radMethod->out1 = comment;
	radMethod->out2 = debug1;
	radMethod->out3 = debug2;

	matrixPane.SetBgColour(bgColour);
	matrixScene = slBeginObject("matrix");
	slObject(new scMatrixDisplay((Void **) &radMethod));
	slEndObject();
	matrixPane.SetScene(matrixScene);
	matrixPane.Redraw();
}

Void RadForm::SetupMenus()
{
	// Set up the menus

	fl_set_choice_fontsize(methodChoice, FL_SMALL_SIZE);
	fl_set_choice_fontstyle(methodChoice, FL_BOLD_STYLE);
	fl_set_choice_fontsize(basis, FL_SMALL_SIZE);
	fl_set_choice_fontstyle(basis, FL_BOLD_STYLE);
	fl_set_choice_fontsize(transPanel.visMethod, FL_TINY_SIZE);
	fl_set_choice_fontsize(transPanel.quadLevel, FL_TINY_SIZE);
}

Void RadForm::SetupControls()
{
	// Set values for the various UI components
			
	fl_set_slider_bounds(patchSubdivs, 0.0, 20);	
	fl_set_slider_value(patchSubdivs, 1.0);	
	fl_set_slider_return(patchSubdivs, 0);	
	
	fl_set_slider_bounds(eltSubdivs, 0.2, 20);	
	fl_set_slider_value(eltSubdivs, 0.5);
	fl_set_slider_return(eltSubdivs, 0);
	
	fl_set_slider_bounds(alphaLevel, 1, 2);	
	fl_set_slider_value(alphaLevel, 1);	
	fl_set_slider_return(alphaLevel, 0);
	
	fl_set_slider_bounds(error, -7.0, 0.0);
	fl_set_slider_step(error, 0.1);
	fl_set_slider_precision(error, 4);
	fl_set_slider_filter(error, esliderFilt);
	fl_set_slider_value(error, -3.0);

	fl_set_slider_bounds(kFError, -7.0, 0.0);
	fl_set_slider_step(kFError, 0.1);
	fl_set_slider_precision(kFError, 4);
	fl_set_slider_filter(kFError, esliderFilt);
	fl_set_slider_value(kFError, -2.0);
	
	fl_set_slider_bounds(kAError, -7.0, 2.0);
	fl_set_slider_step(kAError, 0.1);
	fl_set_slider_precision(kAError, 4);
	fl_set_slider_filter(kAError, esliderFilt);
	fl_set_slider_value(kAError, -6.0);

	fl_set_counter_bounds(plotRes, 1, 32);
	fl_set_counter_step(plotRes, 1, 1);
	fl_set_counter_value(plotRes, 6);
	
	fl_set_counter_bounds(renderPanel.maxIter, 1, 16);
	fl_set_counter_step(renderPanel.maxIter, 1, 1);
	fl_set_counter_bounds(renderPanel.bestLevels, -1, 8);
	fl_set_counter_step(renderPanel.bestLevels, 1, 1);

	my_deactivate(examRad);		// Deactivate plot buttons.
	my_deactivate(examFF);
}

#include "SetGet.cc"

Void RadForm::SampleFF(Matd &samples)
// Sample the form factor values from the currently selected source patch over
// the currently selected destination patch.
{
	Int			i, j;
	RadElem		*to = mainPane.selPatch[0];
	RadElem		*from = mainPane.selPatch[1];
	Coord		place;
	Vector		x, y;
	Point		point;
	Int			density;
	
	if (options.method == kHierarchical && 
			(options.basis == kFlatlet3 || options.basis == kMultiwavelet3))
		density = (Int) (fl_get_counter_value(plotRes) * 3);
	else
		density = (Int) (fl_get_counter_value(plotRes) * 4);
	
	samples.SetSize(density + 1, density + 1);
	x = to->Vertex(2) - to->Vertex(1);
	y = to->Vertex(0) - to->Vertex(1);
		
	for (i = 0; i <= density; i++)
		for (j = 0; j <= density; j++)
		{
			if (i == 0)
				place[0] = 1e-6;
			else
				place[0] = i / GCLReal(density) - 1e-6;

			if (j == 0)
				place[1] = 1e-6;
			else
				place[1] = j / GCLReal(density) - 1e-6;
			
			point = x * place[0] + y * place[1] + to->Vertex(1);
			samples[i][j] = 
				from->ApproxPatchFactor(point, to->Normal());
		}
}

Void RadForm::SamplePatch(Matd &samples)
// Sample the currently selected patch.
{
	Int			i, j;
	Coord		place;
	Int			density;
	
	if (options.method == kHierarchical && (options.basis == kFlatlet3 ||
											options.basis == kMultiwavelet3))
		density = (Int) fl_get_counter_value(plotRes) * 9;
	else
		density = (Int) fl_get_counter_value(plotRes) * 8;

	samples.SetSize(density, density);
		
	for (i = 0; i < density; i++)
		for (j = 0; j < density; j++)
		{
			place[0] = GCLReal(i) / GCLReal(density - 1);
			place[1] = GCLReal(j) / GCLReal(density - 1);
						
			samples[i][j] =  len((mainPane.selPatch[0])->Sample(place));
		}
}


Void RadForm::UpdateMethodOptions(Int oldMethod, Int newMethod)
{
	switch (oldMethod)
	{
	case kMatrix:
		my_deactivate(alphaLevel);
		my_deactivate(patchSubdivs);
		break;
	case kProgressive:
		my_deactivate(alphaLevel);
		my_deactivate(patchSubdivs);
		my_deactivate(error);
		break;
	case kProgSubstruct:
		my_deactivate(error);
		my_deactivate(alphaLevel);
		my_deactivate(eltSubdivs);
		my_deactivate(patchSubdivs);
		my_deactivate(kFError);
		my_deactivate(kAError);
		break;
	case kHierarchical:
		my_deactivate(alphaLevel);
		my_deactivate(basis);
		my_deactivate(error);
		my_deactivate(kFError);
		my_deactivate(kAError);
		break;
	case kAnalytical:
		my_deactivate(patchSubdivs);
		break;
	}
		
	switch (newMethod)
	{
	case kMatrix:
		my_activate(alphaLevel);
		my_activate(patchSubdivs);
		break;
	case kProgressive:
		my_activate(alphaLevel);
		my_activate(patchSubdivs);
		my_activate(error);
		break;
	case kProgSubstruct:
		my_activate(error);
		my_activate(alphaLevel);
		my_activate(eltSubdivs);
		my_activate(patchSubdivs);
		my_activate(kFError);
		my_activate(kAError);
		break;
	case kHierarchical:
		my_activate(alphaLevel);
#ifdef RAD_HO
		my_activate(basis);
#else
		my_deactivate(basis); 
#endif
		my_activate(error);
		my_activate(kFError);
		my_activate(kAError);
		break;
	case kAnalytical:
		my_activate(patchSubdivs);
	}
}

Void RadForm::RenderMode(Bool on)
{
	if (!on)
	{
		UpdateMethodOptions(kNoMethod, options.method);
		my_activate(goBtn);
		my_activate(avar);
		my_activate(exitBtn);
		my_activate(methodChoice);
	}
	else
	{
		UpdateMethodOptions(options.method, kNoMethod);
		my_deactivate(goBtn);
		my_deactivate(avar);
		my_deactivate(exitBtn);
		my_deactivate(methodChoice);
	}
}

scScenePtr RadForm::LoadScene()
{
	const Char *result;

	result = fl_show_fselector("Please pick a scene file", 0, "*.*", 0);
	return(LoadScene(result));
}

scScenePtr RadForm::LoadScene(const Char *path)
{
	scScenePtr	scene = 0;
	FileName	filename;
	
	if (path)
	{
		filename.SetPath(path);
		scene = SceneReader::Load(filename);
		if (scene)
			scene->Normalise();
	}

	if (!scene)
		scene = BoxScene();
		
	return(scene);
}

// --- Event Handling ---------------------------------------------------------

static const Char *gHelpURL = 
		"http://www.cs.cmu.edu/~radiosity/radvis-help.html";

Void RadForm::StartHelp()
{
	String		line;
	Int			err;
	StatusForm	status;
	Char		*netscape = "netscape";
	
	// Fire up nutscape, if necessary, and target the help URL.
	// XXX Should change this to use exec*() and wait()...
	
	status.Show();
	Field(status.heading) << "Help" << show;
	Field(status.body) << "Opening " << gHelpURL << show;
	line.Printf("%s -remote 'openURL(%s, new-window)'", netscape, gHelpURL);
	err = system(line);

	if (err)
	{
		Field(status.body) << "remote failed (" << err << ")" << show;
		line.Printf("%s %s&", netscape, gHelpURL);
		Field(status.body) << "Starting Netscape..." << show;
		err = system(line);
		if (err)
			cout << "Could not invoke Netscape: help is available at "
				 << gHelpURL << show;
		else
			Field(status.body) << "All systems go..." << show;
	}
	else
		Field(status.body) << "All systems go." << show;
	
	sleep(2);
	status.Hide();
}

Void RadForm::UpdateRadScene()
{
	GetFormOptions();

	Int saveWire = options.wire;
	options.wire = true;
	if (!sceneSet)
	{
		radMethod->SetScene(scene);
		mainPane.SetHeadlight(false);
		mainPane.SetScene(radMethod->GetScene());
	}
	else
		radMethod->ResetOptions();

	mainPane.Redraw();
	options.wire = saveWire;
	sceneSet = true;

	SetFormOptions();
}

Void RadForm::ChooseMethod()
{
	// Change the selected radiosity method
	
	if (oldMethodChoice != fl_get_choice(methodChoice))
	{
		delete radMethod;
		UpdateMethodOptions(oldMethodChoice, fl_get_choice(methodChoice));

		GetFormOptions();
		radMethod = RadMethod::NewRadMethod();
		radMethod->SetDisplay(&mainPane, fgs);
		radMethod->matDisplay = &matrixPane;
		radMethod->out1 = comment;
		radMethod->out2 = debug1;
		radMethod->out3 = debug2;

		if (options.method >= kHierarchical)
			options.visibility = vis_4x4;
		else
			options.visibility = vis_1;

		sceneSet = false;
		Reset();
		oldMethodChoice = fl_get_choice(methodChoice);
	}
}

Void RadForm::Reset()
{
	// reset the current radiosity run
		
	options.stop = true;
	options.step = false;
	options.pause = false;
	options.pvData = 0;
	my_deactivate(examRad);		// Deactivate plot buttons.
	my_deactivate(examFF);
	fl_set_button(pause, options.pause);

	// go back to the original scene.

	sceneSet = false;
	mainPane.SetHeadlight(true);
	mainPane.SetScene(scene);
	avarPanel.scenePane = 0;
	Field(comment) << "Current scene: " << scene->Label() << show;
	SetFormOptions();
}

Void RadForm::Go()
{
	// Render the current scene.
				
	RenderMode(true);
	options.stop = false;

	// Make sure the scene has been attached to the renderer.

	UpdateRadScene();
//	avarPanel.Hide(); XXX need to fix xforms error msg before we can
//	do this.

	// Doooo it...

	radMethod->Render();

	RenderMode(false);
}

Void RadForm::SceneSelect()
{
	// we default to showing the matrix, and no volume clustering
	// for simple scenes. 
	options.drawMatrix = true;
	options.cluster = false;

	switch(fl_get_choice(sceneSelect))
	{
	case menu_box:
		scene = BoxScene(); break;
	case menu_parallel:
		scene = ParallelScene(); break;
	case menu_sidelight:
		scene = SidelightScene(); break;
	case menu_blocker:
		scene = BlockerScene(); break;
	case menu_abutting:
		scene = AbuttingScene(); break;
	case menu_table:
		scene = TableScene();
		options.drawMatrix = false;
		options.cluster = true;
		break;
	case menu_tetra:
		scene = BoxScene(1); break;
	case menu_load:
		scene = LoadScene(); 
		options.drawMatrix = false;
		options.cluster = true;
		break;
	}
		
	sceneSet = false;
	Reset();
}

Void RadForm::DumpImage()
{
	const Char	*result;
	RGBAImage	image;
	FileName	imageFile;
	
	result = fl_show_fselector("Saving main image...", 0, "*.tiff,*.tif,*.ppm",
							   "radvis.tif");
	mainPane.Redraw();

	if (result)
	{
		imageFile.SetPath(result);
		imageFile.MakeUnique();
		cout << "*** dumping image to " << result << endl;

#ifdef GCL_MESA
		MesaRenderer 	mr;
		RGBAImage		hqImage;
		Int				xsize = 400, ysize = 400;
				
		image.SetSize(2 * xsize, 2 * ysize);
		hqImage.SetSize(xsize, ysize);
		mr.Init(image);
		mr.SetHeadlight(false);
		mr.SetDoubleSided(true);
		mr.SetBgColour(bgColour);
		mr.Clear().Draw(mainPane.ItsScene()).Show();
		image.DownSample(hqImage);

		hqImage.Save(imageFile);
#else
		mainPane.GetImage(image);
		image.Save(imageFile);
#endif
	}	
}

Void RadForm::DumpObjFile()
{
	const Char	*result;

	result = fl_show_fselector("Saving wavefront file...", 0, "*.obj,*.obj.gz",
							   "radvis.obj");
	mainPane.Redraw();

	if (result)
	{
		cout << "*** dumping wavefront file to " << result << endl;

		radMethod->WriteObjFile(FileName().SetPath(result).MakeUnique().GetPath());
	}	
}

Void RadForm::DumpSLFile()
{
	const Char	*result;

	result = fl_show_fselector("Saving SL file...", 0, "*.sl",
							   "out.sl");
	mainPane.Redraw();

	if (result)
	{
		cout << "*** dumping GCL 'SL' file to " << result << endl;

		radMethod->WriteSLFile(result);
	}	
}

Void RadForm::DumpEPS()
{
	const Char	*result;
	ofstream	fout;

	result = fl_show_fselector("Saving main image...", 0, "*.eps", "radvis.eps");

	if (result)
	{
		EPSRenderer eps;
		cout << "*** dumping eps file to " << result << endl;
			
		eps.Attach(FileName().SetPath(result).MakeUnique().GetPath());
		eps.Clear().Draw(mainPane.ItsScene()).Show();
	}			
}

Void RadForm::DumpCamera()
{
	const Char *result;
	ofstream	fout;
		
	result = fl_show_fselector(
		"Saving camera file...", 0, "*.cam", "out.cam");
		
	if (result)
	{
		cout << "*** saving camera to " << result << endl;
		fout.open(result);
		if (fout)
			((scCamera*) mainPane.ItsCamera())->Print(fout);
		else
			perror("Couldn't open output file");
	}


}

Void RadForm::Event(FL_OBJECT *object)
{
	// Handle UI events. This should be split out into (many) different
	// routines!

	if (object == exitBtn)
	{
		// Our work here is done.		
		fgs->SignalDone();
	}
	else if (object == goBtn)
		Go();
	else if (object == methodChoice)
		ChooseMethod();
	else if (object == resetBtn)
		Reset();
	else if (object == step)
	{
		options.step = true;
		options.pause = true;
   		fl_set_button(pause, options.pause);
	}
	else if (object == sceneSelect)
		SceneSelect();
	else if (object == save)
	{
		switch(fl_get_menu(save))
		{
		case menu_savePic:
			DumpImage();
			break;
		case menu_saveEPS:
			DumpEPS();
			break;
		case menu_saveSL:
			DumpSLFile();
			break;
		case menu_saveObj:
			DumpObjFile();
		}
	}
	else if (
		object == gouraudBtn ||
		object == textureBtn ||
		object == wireFrame || 
		object == patchView ||
		object == funcView || 
		object == anchor ||
		object == animation 
		)
	{
		// These are option-changing controls: we update the options and redraw
		// the scene in case it or the way of drawing it has changed.
	
		GetFormOptions();
		mainPane.Redraw();	

	}
	if (object == methodOpts)
	{
		Int item = fl_get_menu(methodOpts);
		if (item == menu_moreOpts)
		{	
			renderPanel.Show();
			return;
		}

		GetFormOptions();
	}
	else if (
			object == patchSubdivs ||
			object == eltSubdivs
		)
	{
		// notify rad object that meshing parameters have changed
		UpdateRadScene();
	}
	else if (object == avar)
	{
		sceneSet = false;
		Reset();
		
		//	Show the avar panel
		
		avarPanel.Show();
		if (avarPanel.scenePane == 0)
		{
			avarPanel.SetScene(scene);
			avarPanel.scenePane = &mainPane;
			scene->Set(avarPanel.avarList);
			avarPanel.Init();
		}
	}
	else if (object == examRad && mainPane.selPatch[0])
	{
		Examiner *newPlot = new Examiner;
		
		newPlot->Show();
		newPlot->Init(*fgs, &mainPane);
		SamplePatch(newPlot->plot->samples);
		newPlot->plotPane.SetBgColour(bgColour);
		newPlot->plotPane.Redraw();
	}
	else if (object == examFF && mainPane.selPatch[0] && mainPane.selPatch[1])
	{
		Examiner *newPlot = new Examiner;
		
		newPlot->Show();
		newPlot->Init(*fgs, &mainPane);
		newPlot->plotPane.SetBgColour(bgColour);
		SampleFF(newPlot->plot->samples);
		newPlot->plotPane.Redraw();
	}
	else if (object == ffShow)
		transPanel.Show();
	else if (object == msgPick)
	{
		if (mainPane.selPatch[0])
			my_activate(examRad);
		else
			my_deactivate(examRad);
			
		if (mainPane.selPatch[0] && mainPane.selPatch[1])
			my_activate(examFF);
		else
			my_deactivate(examFF);

		matrixPane.Redraw();
	}
	else if (object == help)
		StartHelp();
	else
		GetFormOptions();
}


// --- Options panel stuff ----------------------------------------------------


Void TransportPanel::Event(FL_OBJECT *object)
{
	// nice & simple -- just hide the form if 'OK' is
	// clicked.

	if (object == dismiss)
		Hide();
	else 
		parent->GetFormOptions();
};

Void RenderPanel::Event(FL_OBJECT *object)
{
	// nice & simple -- just hide the form if 'OK' is
	// clicked.

	if (object == dismiss)
		Hide();
	else 
		parent->GetFormOptions();
};


// --- Main -------------------------------------------------------------------


Void RadForm::InitOptions(Int argc, Char **argv, RadControl &options, 
						  Char **scenePathP)
{
	Int         ndm, meshRnd, meshNonLin, tri1, tri2, vers, help, noFork;  
	ArgForm    *arg_format;       
	Double		grey, clr[3];
	Double		meshComp = 0.1;
	
	grey = -1.0;
	clr[0] = 0.4; clr[1] = 0.5; clr[2] = 0.6;

	arg_format = arg_to_form(0,
		"[%S]",			scenePathP,				"Specify scene file",
		"-v",			ARG_FLAG(&vers),		"Print version & object info",
		"-h",		ARG_FLAG(&help),			"Help",
		"-noMatrix",	ARG_FLAG(&ndm),			"Turn off matrix display",
		"-bgr %F",		&grey,		 			"Set background grey level",
		"-bgc %F %F %F", clr, clr + 1, clr + 2,	"Set colour background",
		"-meshComp %F",	&meshComp,			 	"Specify initial mesh complexity",
		"-meshRandom",	ARG_FLAG(&meshRnd), 	"Vary the mesh density",
		"-meshNonLin",	ARG_FLAG(&meshNonLin),	"Mesh more heavily at edges",
		"-tri1",		ARG_FLAG(&tri1),		"Triangulate built-ins (1)",
		"-tri2",		ARG_FLAG(&tri2),		"Triangulate built-ins (2)",
		"-noFork",		ARG_FLAG(&noFork),		"Don't run in the background",
		0
		);

	if (arg_parse_argv(argc, argv, arg_format) < 0 || help)
	{
		String	usage;
		
		usage.Printf("Usage: %s [options], where options are as follows:", 
				argv[0]);

		fprintf(stderr, "RadVis version %s\n%s\n\n%s\n\n",
				RadGetVersion().CString(),
				RadGetCopyright().CString(),
				usage.CString()
			);
		arg_form_print(arg_format);
		cerr << endl;
		SceneReader::PrintSupportedFormats(cerr);
		exit(1);
	}

	if (vers)
	{
		cout << "RadVis version " << RadGetVersion() << endl;
		cout << "Radiosity visualisation, " << RadGetCopyright() << endl;
		exit(1);
	}

#ifndef DEBUG
	if (!noFork && fork())
		exit(0);
#endif

	if (tri1)
		gQuadType = 1;
	else if (tri2)
		gQuadType = 2;
		
	options.drawMatrix = !ndm;
	if (meshRnd)
		options.mesh = mesh_random;
	if (meshNonLin)
		options.mesh = mesh_nonlin;

	scMRModel::SetComplexity(meshComp);

#ifdef RAD_MINE
	// radvis-specific
	options.method = kHierarchical;
	options.visibility = vis_4x4;
#endif

	if (grey >= 0.0)
		bgColour = cWhite * grey;
	else if (clr[0] >= 0.0)
		bgColour = Colour(clr[0], clr[1], clr[2]);
}

Int RadForm::Main(Int argc, Char **argv)
{
	Char		*scenePath = 0;

	InitOptions(argc, argv, options, &scenePath);

	Show();	
	Init(scenePath);

	fgs->Run();

	return(0);
}

main(Int argc, Char **argv)
{
	// 'cause this calls fl_initialise, it must be created before...
	FormsGraphicsSystem fgs(argc, argv);
	// the radForm object.
	RadForm		radForm;
	radForm.fgs = &fgs;

	// let's boogie.
	return(radForm.Main(argc, argv));
}
