/*
	File:		WaveSupport.cc

	Function:	Support utilities for the wavelet HElems.

	Author:		Andrew Willmott

	Notes:		
*/

#include "WaveSupport.h"
#include "RadControl.h"

// --- Misc. utilities --------------------------------------------------------

GCLReal ClrVecCalcError(ClrVec B1, ClrVec B2, Int n)
{
	Colour csum(vl_0);
	
	for (int i = 0; i < n; i++)
		csum += *B1++ - *B2++;
	return(dot(csum, kRadRGBToLum) / n);
}

Colour ClrVecAverage(ClrVec B, Int n)
{
	Colour csum(vl_0);
	
	for (int i = 0; i < n; i++)
		csum += *B++;
	return(csum / n);
}

Void ClrVecScale(ClrVec c, Int n, GCLReal s)
{
	for (int i = 0; i < n; i++)
		*c++ *= s;
}

Void ClrVecMult(const GCLMat &m, ClrVec src, ClrVec dst)
// dst = m * src
{
	for (int i = 0; i < m.Rows(); i++)
	{
		dst[i] = vl_0;
		for (int j = 0; j < m.Cols(); j++)
			dst[i] += m[i][j] * src[j];
	}
}

Void ClrVecMultAcc(const GCLMat &m, ClrVec src, ClrVec dst)
// dst += m * src
{
	for (int i = 0; i < m.Rows(); i++)
		for (int j = 0; j < m.Cols(); j++)
			dst[i] += m[i][j] * src[j];
}

// --- Routines for reconstructing a function basis ---------------------------

Colour M2QuadSample(Coord &c, ClrVec B)
{
	Colour result;
	result = (B[0] + B[1] * c[0]) +
			 (B[2] + B[3] * c[0]) * c[1];
	return(result);
}

Colour M2TriSample(Coord &c, ClrVec B)
{
	Colour result;
	result = B[0] * (1 - c[0] - c[1]) + B[1] * c[0] + B[2] * c[1]; 
	return(result);
}

Colour M3QuadSample(Coord &c, ClrVec B)
{
	Colour	result;
	GCLReal	x1 = c[0], x2 = sqr(c[0]) - 1.0 / 3.0;
	GCLReal y1 = c[1], y2 = sqr(c[1]) - 1.0 / 3.0;

	result =      (B[0] + B[1] * x1 + B[2] * x2) +
			 y1 * (B[3] + B[4] * x1 + B[5] * x2) +
			 y2 * (B[6] + B[7] * x1 + B[8] * x2);

	return(result);
}

Colour M3TriSample(Coord &c, ClrVec B)
{
	Colour	result;
	GCLReal ga = 1 - c[1] - c[0];
	GCLReal gb = c[0];
	GCLReal gc = c[1];
	
	result =	B[0] * sqr(ga) + 
				B[1] * sqr(gb) +
				B[2] * sqr(gc) + 
				2 * (
					B[3] * ga * gb +
					B[4] * gb * gc + 
					B[5] * gc * ga
				);

	return(result);
}

Void DrawWaveLink(Renderer &r, GCLReal left, GCLReal top, 
				  GCLReal right, GCLReal bottom,
				  Colour weight, const GCLMat &m)
{
	GCLReal	dx = (right - left) / m.Cols();
	GCLReal dy = (bottom - top) / m.Rows();
	GCLReal oldLeft = left;
	GCLReal *tc = m.Ref();
	Int		i, j;
	
	right = left + dx;
	bottom = top + dy;
	
	for (i = 0; i < m.Rows(); i++)
	{
		for (j = 0; j < m.Cols(); j++)
		{
			r.SetColour(weight * *tc);
			r.Rect(left, top, right, bottom);
			left += dx; right += dx;
			tc++;
		}
		left = oldLeft; right = oldLeft + dx;
		top += dy; bottom += dy;
	}
}

// --- Colour tensor routines -------------------------------------------------

/*
	WARNING: All of the routines below are to deal with applying 1D
	basis transformations in the 2D realm of our quad patches.

	If you are trying to understand the wavelet code, it is probably
	best to start with the triangular wavelets, and ignore the tensor
	stuff at first.
*/

Void ClrVecTensorMult(const GCLMat &m, ClrVec src, ClrVec dst)
// [dst] = m * [src] * trans(m)
// [src] is the matrix S[i][j] = src[i.n + j], where n is the # of
// rows (and columns) in m.
// A tensor-product multiplication is O(n ^ 3/2) where the equivalent full 
// multiplication (using Mult) would be O(n^2), which is why we go to
// this extra trouble to perform it.
{
	Colour	t[16];
	
	ClrVecLeftTensorMult(m, src, t);
	ClrVecRightTensorMult(m, t, dst);
}

Void ClrVecLeftTensorMult(const GCLMat &m, ClrVec src, ClrVec dst)
// dst = m * [src]
{
	for (int i = 0; i < m.Rows(); i++)
		for (int j = 0; j < m.Cols(); j++)
		{
			*dst = vl_0;
			for (int k = 0; k < m.Cols(); k++)
				*dst += m[i][k] * src[k * m.Cols() + j];
			dst++;
		}
}

Void ClrVecRightTensorMult(const GCLMat &m, ClrVec src, ClrVec dst)
// [dst] = [src] * trans(m)
{
	for (int i = 0; i < m.Cols(); i++)
		for (int j = 0; j < m.Cols(); j++)
		{
			*dst = vl_0;
			for (int k = 0; k < m.Cols(); k++)
				*dst += m[j][k] * src[i * m.Cols() + k];
			dst++;
		}
}

Void ClrVecLeftTensorMultAcc(const GCLMat &m, ClrVec src, ClrVec dst)
// [dst] += m * [src]
{
	for (int i = 0; i < m.Rows(); i++)
		for (int j = 0; j < m.Cols(); j++)
		{
			for (int k = 0; k < m.Cols(); k++)
				*dst += m[i][k] * src[k * m.Cols() + j];
			dst++;
		}
}

Void ClrVecRightTensorMultAcc(const GCLMat &m, ClrVec src, ClrVec dst)
// [dst] += [src] * trans(m)
{
	for (int i = 0; i < m.Cols(); i++)
		for (int j = 0; j < m.Cols(); j++)
		{
			for (int k = 0; k < m.Cols(); k++)
				*dst += m[j][k] * src[i * m.Cols() + k];
			dst++;
		}
}




// --- Tensor product matrices ------------------------------------------------

Void MakeTensorMatrix(const GCLMat &c, GCLMat &C)
// Calculates C, where C is the tensor-product of matrix c with itself. 
// O(n^4)
// (Useful because if c governs a 1D transformation, C governs the same
// transformation in 2D.)
{
	Int		n = c.Rows(), n2 = sqr(n);
	Int		i, j, k, l, s, t;
	
	C.SetSize(n2, n2);

	for (i = 0, s = 0; i < n; i++)
		for (j = 0; j < n; j++, s++)
			for (k = 0, t = 0; k < n; k++)
				for (l = 0; l < n; l++, t++)
					C[s][t] = c[j][l] * c[i][k];
}

/* 
   The straightforward way to use 1D rules to do (say) a radiosity transfer
   would be to use MakeTensorMatrix() to generate 2D versions of R and S
   from the corresponding 1D versions, and then calculate T = R' * K * S'.
   
   However, it's faster to not calculate R' and S' explicitly, and instead
   use the tensor multiplication routines below.
*/

Void MatRightTensorMult(const GCLMat &a, const GCLMat &c, GCLMat &d)
// d = a * C, where C is the tensor-product of matrix c with itself.
// O(mn^3).
{
	Int		n = c.Rows();
	Int		i, j, k, l, s, t, m;
	GCLMat	p;
	
	d.SetSize(a.Rows(), sqr(n));
	d = vl_0;
	p.SetSize(c);
	
	for (m = 0; m < a.Rows(); m++)
	{
		p = vl_0;
		for (l = 0; l < n; l++)
			for (i = 0, s = 0; i < n; i++)
				for (j = 0; j < n; j++, s++)
					p[i][l] += c[j][l] * a[m][s];
					
		for (k = 0, t = 0; k < n; k++)
			for (l = 0; l < n; l++, t++)
				for (i = 0; i < n; i++)
					d[m][t] += c[i][k] * p[i][l];
	}
}

Void MatLeftTensorMult(const GCLMat &c, const GCLMat &a, GCLMat &d)
// d = C * a, where C is the tensor-product of matrix c with itself.
// O(mn^3).
{
	Int		n = c.Rows();
	Int		i, j, k, l, s, t, m;
	GCLMat	p;
	
	d = vl_0;
	p.SetSize(c);
		
	for (m = 0; m < a.Cols(); m++)
	{
		p = vl_0;
		for (j = 0; j < n; j++)
			for (k = 0, t = 0; k < n; k++)
				for (l = 0; l < n; l++, t++)
					p[j][k] += c[j][l] * a[t][m];
					
		for (i = 0, s = 0; i < n; i++)
			for (j = 0; j < n; j++, s++)
				for (k = 0; k < n; k++)
					d[s][m] += c[i][k] * p[j][k];
	}
}

