/*
	File:		SampleView.cc

	Function:	View sample patterns generated by samplegen

	Author:		Andrew Willmott

	Notes:		
*/

#include "gcl/GLXPane.h"
#include "Samples.h"
#include "gcl/Scene.h"
#include "gcl/Draw.h"
#include "gcl/VecUtil.h"
#include "cl/ArgParse.h"
#include <stdlib.h>
#include <unistd.h>

static XGraphicsSystem *gXGS = 0;
static XEventPane *gPane = 0;

static Bool gStep = false;
static Int gMaxSamples = 1024;
static Int g3D = 0;
static Int gReverse = 0;

UInt32 BitReverseStd(UInt32 in, UInt32 places)
{
	UInt32 out = 0;

	while (places--)
	{
		out <<= 1;
		if (in & 1)
			out |= 1;
		in >>= 1;
	}

	return(out);
}

Int Log2(UInt32 x)
// return m such that 2^m > x
{
	Int r = 32;

	if (!x)
		return(0);
	if (!(x & 0xffff0000))
	{
		x <<= 16;
		r -= 16;
	}
	if (!(x & 0xff000000))
	{
		x <<= 8;
		r -= 8;
	}
	if (!(x & 0xf0000000))
	{
		x <<= 4;
		r -= 4;
	}
	if (!(x & 0xc0000000))
	{
		x <<= 2;
		r -= 2;
	}
	if (!(x & 0x80000000))
		r -= 1;

	return(r);
}

Int Log2Ceil(UInt32 x)
// return m such that 2^m >= x
{
	Int r = 32;

	if (!x)
		return(0);
	if (!(x & 0xffff0000))
	{
		x <<= 16;
		r -= 16;
	}
	if (!(x & 0xff000000))
	{
		x <<= 8;
		r -= 8;
	}
	if (!(x & 0xf0000000))
	{
		x <<= 4;
		r -= 4;
	}
	if (!(x & 0xc0000000))
	{
		x <<= 2;
		r -= 2;
	}
	if (!(x & 0x80000000))
	{
		x <<= 1;
		r -= 1;
	}
	if (!(x & 0x7fffffff))
		r -= 1;

	return(r);
}


/*
 * inside_out: turn a number "inside-out": a generalization of bit-reversal.
 * For n = power of two, this is equivalent to bit-reversal.
 *
 * Turn the number a inside-out, yielding b.  If 0<=a<n then 0<=b<n.
 * Also return r = min(n-b, largest power of 2 dividing b)
 */

Int BitReverse(Int n, Int a)
{
	Int k, m, b;

	m = n;

	for (b = 0, k = 1; k < n; k <<= 1)
		if (a << 1 >= m)
		{
			b += k;
			a -= (m + 1) >> 1;
			m >>= 1;
		}
		else
			m = (m + 1) >> 1;

	return(b);
}


Colour FindSampleColour(Int i, Int n)
{
	if (gReverse)
		i = BitReverse(n, i);

	return(HSVCol(360.0 * i / n, 1.0, 1.0));
}

Void DoStep(Renderer &r)
{
	if (gStep)
	{
		cout << "hit escape to continue" << endl;
		gXGS->Run();
		r.Show();
		gPane->HandleExpose();
	}
}


Void DrawSquareSamples16(Renderer &r)
{
	Int		i, j;

	r.Push();
	r.Xform(Shift(Vector(-1, -1, -1))).Xform(Scalef(2.0));
	r.C(cWhite).Begin(renPoly).P(Point(0, 0, 0)).P(Point(1, 0, 0))
		.P(Point(1, 1, 0)).P(Point(0, 1, 0)).End();

	for (j = 0; j < Min(gMaxSamples, kNumSampArrays); j++)
	{
		r.C(FindSampleColour(j, kNumSampArrays));

		for (i = 0; i < 16; i++)
		{
			DrawCircle(r, Point(tSquareSamples16[j][i][0], 
								tSquareSamples16[j][i][1], 0.01), vl_z, 0.01);
			DoStep(r);
		}
	}

	r.Pop().Show();
}

Void DrawTriSamples16(Renderer &r)
{
	Int		i, j;

	r.Push();
	r.Xform(Shift(Vector(-1, -1, -1))).Xform(Scalef(2.0));
	r.C(cWhite).Begin(renPoly).P(Point(0, 0, 0)).P(Point(1, 0, 0))
		.P(Point(0, 1, 0)).End();

	for (j = 0; j < Min(gMaxSamples, kNumSampArrays); j++)
	{
		r.C(FindSampleColour(j, kNumSampArrays));

		for (i = 0; i < 16; i++)
		{
			DrawCircle(r, Point(tTriangleSamples16[j][i][0], 
								tTriangleSamples16[j][i][1], 0.01),
					   vl_z, 0.01);
			DoStep(r);
		}
	}

	r.Pop().Show();
}

Void DrawCubeView(Renderer &r, const Vector &dir, Bool ortho = true)
{
	Int		i;

	r.Xform(Shift(Vector(-1, -1, -1))).Xform(Scalef(2.0));

	r.C(cWhite);

	if (ortho)
	{
		Point	p1(vl_0), p2(vl_1);
		Int		pi = MaxEltIndex(dir);

		p1[pi] = p2[pi] = -0.01;
		PaintBox(r, p1, p2);
	}
	else
		FrameBox(r, vl_0, vl_1);
		
	for (i = 0; i < Min(kMaxStratSamples, gMaxSamples); i++)
	{
		r.C(FindSampleColour(i, kMaxStratSamples));

		DrawCircle(r, Point(tCubeSamples[i][0], 
			tCubeSamples[i][1], tCubeSamples[i][2]), dir, 0.01);

		DoStep(r);
	}
}

Void DrawSphereView(Renderer &r, const Vector &dir, Bool ortho = true)
{
	Int		i;
	Double	rad;
	
	r.C(cWhite);
	scRenderSteps = 32;

	if (ortho)
	{
		Point p(vl_0);

		p[MaxEltIndex(dir)] = -1.0;
		DrawCircle(r, p, norm(dir), 1.0);
	}
	else
	{
		rad = sqrlen(dir);
		rad = sqrt(rad / (rad - 1));
		scPrimitive::sRenderStyle = renLineLoop;
		DrawCircle(r, Point(vl_0), norm(dir), rad);
		scPrimitive::sRenderStyle = renPoly;
	}

	scRenderSteps = 8;

	for (i = 0; i < Min(kMaxStratSamples, gMaxSamples); i++)
	{
		r.C(FindSampleColour(i, kMaxStratSamples));

		DrawCircle(r, Point(tSphereSamples[i][0], 
			tSphereSamples[i][1], tSphereSamples[i][2]), dir, 0.02);

		DoStep(r);
	}
}

Void Draw3DSamples(Bool sphere, Renderer &r)
{
	r.Push();

	if (g3D)
	{
		SceneCamera	cam;
		Point vp(3.0, 3.0, 4.0);
		
		cam.SetFOV(22.5);
		cam.SetLookAt(vp, vl_0);
		r.Cam(cam);

		if (sphere)
		{
			r.Xform(Scalef(1.5));
			DrawSphereView(r, vp / 1.5, false);
		}
		else
			DrawCubeView(r, vp.Normalise(), false);
	}
	else
	{
		r.Xform(Scalef(0.45)).Push();
		r.Xform(Shift(-2.1, 0, 0)).Xform(Rotation(vl_y, -vl_halfPi));
		if (sphere)
			DrawSphereView(r, vl_x);
		else
			DrawCubeView(r, vl_x);
		r.Pop().Push();
		r.Xform(Rotation(vl_x, vl_halfPi));
		if (sphere)
			DrawSphereView(r, vl_y);
		else
			DrawCubeView(r, vl_y);
		r.Pop().Push();
		r.Xform(Shift(2.1, 0, 0));
		if (sphere)
			DrawSphereView(r, vl_z);
		else
			DrawCubeView(r, vl_z);
		r.Pop();
	}

	r.Pop().Show();
}

Void DrawSquareSamples(Renderer &r)
{
	Int		i;

	r.Push();
	r.Xform(Shift(Vector(-1, -1, -1))).Xform(Scalef(2.0));
	r.C(cWhite).Begin(renPoly).P(Point(0, 0, 0)).P(Point(1, 0, 0))
		.P(Point(1, 1, 0)).P(Point(0, 1, 0)).End();

	for (i = 0; i < Min(kMaxStratSamples, gMaxSamples); i++)
	{
		r.C(FindSampleColour(i, kMaxStratSamples));

		DrawCircle(r, Point(tSquareSamples[i][0], 
							tSquareSamples[i][1], 0.01), vl_z, 0.01);
		DoStep(r);
	}

	r.Pop().Show();
}

Void DrawTriSamples(Renderer &r)
{
	Int		i;

	r.Push();
	r.Xform(Shift(Vector(-1, -1, -1))).Xform(Scalef(2.0));
	r.C(cWhite).Begin(renPoly).P(Point(0, 0, 0)).P(Point(1, 0, 0))
		.P(Point(0, 1, 0)).End();

	for (i = 0; i < Min(kMaxStratSamples, gMaxSamples); i++)
	{
		r.C(FindSampleColour(i, kMaxStratSamples));

		DrawCircle(r, Point(tTriangleSamples[i][0], 
							tTriangleSamples[i][1], 0.01),
				   vl_z, 0.01);

		DoStep(r);
	}

	r.Pop().Show();
}

Void DrawCircleSamples(Renderer &r)
{
	Int		i;

	r.Push();

	scRenderSteps = 30;
	DrawCircle(r, vl_0, vl_z, 1.0);

	for (i = 0; i < Min(kMaxStratSamples, gMaxSamples); i++)
	{
		r.C(FindSampleColour(i, kMaxStratSamples));

		DrawCircle(r, Point(tCircleSamples[i][0], 
							tCircleSamples[i][1], 0.01),
				   vl_z, 0.02);

		DoStep(r);
	}

	r.Pop().Show();
}

#ifdef SVIEW_DRAW_PIXEL

Coord BoundaryPoint(Int side)
{
	if ((side & 2) == 0)
		return(Coord(drand48(), side & 1));
	else
		return(Coord(side & 1, drand48()));
}

const Int tRayC1[6] = { 0, 0, 0, 1, 1, 2 };
const Int tRayC2[6] = { 1, 2, 3, 2, 3, 3 };

const Int tStartCorners[12] = 
{ 3, 0, 2, 0, 2, 2, 
  1, 1, 1, 3, 3, 0 };
const Int tCycleLen[12] = 
{ 2, 1, 3, 3, 1, 2,
  2, 3, 1, 1, 3, 2 };

Void DrawPixel(Renderer &r)
// quick hack to draw a pixel covered by 'samps' random polygons
{
	Int		i, numPolys = gMaxSamples, ic, j;
	Coord   corner[4];

	glDisable(GL_DEPTH_TEST);
	srand48(12);
	corner[0] = Coord(0.0, 0.0);
	corner[1] = Coord(1.0, 0.0);
	corner[2] = Coord(1.0, 1.0);
	corner[3] = Coord(0.0, 1.0);

	r.Push();
	r.Xform(Shift(Vector(-1, -1, -1))).Xform(Scalef(2.0));
	r.C(cWhite).Begin(renPoly).P(Point(0, 0, -1)).P(Point(1, 0, -1))
		.P(Point(1, 1, -1)).P(Point(0, 1, -1)).End();

	// pick random half space
	for (i = 0; i < numPolys; i++)
	{
		Int		half = random() % 12, ray = half % 6;
		Coord	c1 = BoundaryPoint(tRayC1[ray]);
		Coord	c2 = BoundaryPoint(tRayC2[ray]);
		GCLReal	depth = Double(i) / numPolys;
		
		r.C(FindSampleColour(i, numPolys));

		r.Begin(renPoly);

		if (half < 6)
			r.P(c1).P(c2);
		else
			r.P(c2).P(c1);
		
		ic = tStartCorners[half];

		for (j = 0; j < tCycleLen[half]; j++)
		{
			r.P(corner[ic]);
			ic = (ic + 1) % 4;
		}	
		r.End();

		DoStep(r);
	}
	
	r.Pop().Show();
}
#endif

Int main(Int argc, Char *argv[])
{
	XGraphicsSystem 	*xgs = new XGraphicsSystem;
	GLXOffscreenPane	*pane = new GLXOffscreenPane();
	XBackedPane			*bpane = new XBackedPane;
	Int					size, noFork, tri, sq, circ, cube, sphere, step, dump;
	Int					t16, s16, xsize;
	Double				grey, clr[3];
	Arg_form			*arg_format;
	SceneCamera			cam;

	// set options

	gXGS = xgs;

	size = 400;
	grey = 0.5;
	clr[0] = -1;

	arg_format = arg_to_form(0,
		"", 									"usage: sview [options]",

		"-square", ARG_FLAG(&sq),				"Draw square samples",
		"-triangle", ARG_FLAG(&tri),			"Draw triangle samples",
		"-circle", ARG_FLAG(&circ),				"Draw circle samples",
		"-cube", ARG_FLAG(&cube),				"Draw cube samples",
		"-sphere", ARG_FLAG(&sphere),			"Draw sphere samples",
		"-s16", ARG_FLAG(&s16),					"Draw magic-square samples",
		"-t16", ARG_FLAG(&t16),					"Draw magic-square samples for triangles",
		"-step", ARG_FLAG(&step),				"Step through samples",
		"-size %d", &size, 						"Size of window",
		"-samps %d", &gMaxSamples, 				"Number of samples to use [1024]",
		"-bgr %F", &grey,						"Set greyscale background",
		"-bgc %F %F %F", clr, clr + 1, clr + 2,	"Set colour background",
		"-dump", ARG_FLAG(&dump), 				"Dump picture",
		"-t", ARG_FLAG(&g3D), 					"Draw cube or sphere in 3D (default is"
												" three \northographic projections)",
		"-r", ARG_FLAG(&gReverse), 				"Colour consecutive samples as differently as possible",
		"-noFork", ARG_FLAG(&noFork), 			"Don't fork into background",

		"", 									"",
		"", "",

		0);

	// Do arg parsing
	if (argc == 1)
	{
		arg_form_print(arg_format);
		exit(0);
	}

	// parse arguments
	if (arg_parse_argv(argc, argv, arg_format) < 0)
		exit(1);

#ifndef DEBUG
	// fork into background
	if (!noFork && fork())
		exit(0);
#endif

	if (step)
		gStep = true;

	cam.SetOrtho(vl_0);
	cam.SetClipping(-2.0, 2.0);

	if (!g3D && (sphere || cube))
	{
		xsize = 2 * size;
		cam.SetAspect(2.0);
	}
	else
		xsize = size;

	xgs->CreateOffscreen(pane, xsize, size);
	if (!dump)
	{
		xgs->CreateWindow(bpane, "SampleView", xsize, size);
		bpane->Attach(pane);
		gPane = bpane;
	}
	
	if (clr[0] < 0)
		pane->SetBgColour(cWhite * grey);
	else
		pane->SetBgColour(Colour(clr[0], clr[1], clr[2]));

	pane->Clear();
	pane->SetCamera(cam);
	pane->SetTransform(Scalef(0.7));

#ifdef SVIEW_DRAW_PIXEL
	DrawPixel(*pane);
#else
	if (circ)
		DrawCircleSamples(*pane);
	else if (sq)
		DrawSquareSamples(*pane);
	else if (tri)
		DrawTriSamples(*pane);
	else if (cube)
		Draw3DSamples(false, *pane);
	else if (sphere)
		Draw3DSamples(true, *pane);
	else if (s16)
		DrawSquareSamples16(*pane);
	else if (t16)
		DrawTriSamples16(*pane);
#endif
		
	if (dump)
	{
		RGBAImage img;
		pane->GetImage(img);
		img.SaveTIFF("out.tiff");
	}
	else
		xgs->Run();
		
	return(0);
}
